import numpy as np
from astropy import units as u
from astropy import constants as const
from amoeba.Classes.blr_streamline import Streamline
from amoeba.Classes.flux_projection import FluxProjection
from amoeba.Util.util import project_blr_to_source_plane


class Torus:

    def __init__(
        self,
        smbh_mass_exp,
        max_height,
        redshift_source,
        radial_step=10,
        height_step=10,
        power_law_density_dependence=0,
        OmM=0.3,
        H0=70,
        **kwargs
    ):
        """Generate a model torus designed to be projected into the plane of the
        accretion disk. Using this column density can be used as an estimate for how
        much light is absorbed by the dust. This is an experimental class.

        Assumes that there is absorption such that:
            flux_abs = flux_emit * (1 - e^{-\rho_{col} / abs_strength})

        where:
            flux_abs is the absorbed flux (an array)
            flux_emit is the emitted flux (an array)
            e is Euler's number
            \rho_{col} is the column density (an array)
            abs_strength is the absorption strength (int/float)

        Note that this can only absorb if there is source flux to absorb. Therefore, the
        FluxProjection of project_extinction_to_source_plane() is cast to an approximate
        magnitude change via -2.5 * np.log10(flux_abs)

        :param smbh_mass_exp: solution to log10(m_smbh / m_sun), which sets
            the size scales of all components in the torus.
        :param max_height: maximum height of the torus in the Z direction, in
            units R_g
        :param redshift_source: redshift of the torus
        :param radial_step: radial resolution in units R_g
        :param height_step: resolution in the Z-axis in units R_g
        :param power_law_density_dependence: defines the density profile as a function
            of radius such that \rho \propto r^{power_law_density_dependence}. Zero means
            the density is assumed to be constant throughout the torus. Positive values
            represent increasing densities as you go further out.
        :param OmM: mass component of the universe's energy budget
        :param H0: Hubble constant in units km/s/Mpc
        """
        self.smbh_mass_exp = smbh_mass_exp
        self.redshift_source = redshift_source
        self.OmM = OmM
        self.H0 = H0
        self.max_height = max_height
        self.radial_step = radial_step
        self.height_step = height_step
        self.max_radius = 0
        self.density_grid = np.zeros(
            (self.max_radius // radial_step + 1, self.max_height // height_step + 1)
        )
        self.radii_values = np.linspace(
            0, self.max_radius, self.max_radius // radial_step + 1
        )
        self.height_values = np.linspace(
            0, self.max_height, self.max_height // height_step + 1
        )

        self.torus_array_shape = np.shape(self.density_grid)
        self.mass = 10 ** (self.smbh_mass_exp) * const.M_sun.to(u.kg)
        self.power_law_density_dependence = power_law_density_dependence

        if (
            "rest_frame_wavelengths" in kwargs.keys()
            and "extinction_coefficients" in kwargs.keys()
        ):
            self.define_extinction_coefficients(
                rest_frame_wavelengths=kwargs["rest_frame_wavelengths"],
                extinction_coefficients=kwargs["extinction_coefficients"],
            )
        else:
            self.define_extinction_coefficients()

    def add_streamline_bounded_region(self, Streamline):
        """Add a streamline representing the boundary of the dust to the torus.

        :param Streamline: Streamline object to use as the inner boundary condition
        :return: True if successful
        """

        assert Streamline.height_step == self.height_step
        assert Streamline.max_height == self.max_height

        new_max_radius = np.sqrt(
            np.max(Streamline.radii_values) ** 2 + np.max(Streamline.height_values) ** 2
        )

        if new_max_radius > self.max_radius:
            self.max_radius = int(new_max_radius)

            dummygrid = np.zeros(
                (
                    self.max_radius // self.radial_step + 1,
                    self.max_height // self.height_step + 1,
                )
            )
            previous_maximum = self.max_radius // self.radial_step + 1
            dummygrid[:previous_maximum, :] = self.density_grid
            self.density_grid = dummygrid

            self.radii_values = np.linspace(
                0, self.max_radius, self.max_radius // self.radial_step + 1
            )

        for hh in range(np.size(self.height_values)):
            mask = self.radii_values >= Streamline.radii_values[hh]
            self.density_grid[np.argmax(mask) :, hh] = 1

        R, Z = np.asarray(
            np.meshgrid(self.radii_values, self.height_values, indexing="ij")
        )
        radii = np.sqrt(R**2 + Z**2)
        self.density_grid *= radii**self.power_law_density_dependence

        self.torus_array_shape = np.shape(self.density_grid)

        return True

    def define_extinction_coefficients(
        self, rest_frame_wavelengths=None, extinction_coefficients=None
    ):
        """Define the extinction coefficients for a list of wavelengths.

        :param rest_frame_wavelengths: list/array of wavelengths with representative
            extinction coefficients.
        :param extinction_coefficients: list/array of extinction coefficients. Must be
            the same shape as rest_frame_wavelengths.
        :return: True if successful
        """
        self.rest_frame_wavelengths = rest_frame_wavelengths
        self.extinction_coefficients = extinction_coefficients

        return True

    def interpolate_to_extinction_at_wavelength(self, observer_frame_wavelength):
        """Use linear interpolation of the extinction curve to determine the extinction
        at a particular wavelength.

        :param observer_frame_wavelength: observer frame wavelength in nm
        :return: int representing the extinction coefficient at the given wavelength
        """

        rest_frame_wavelength = observer_frame_wavelength / (1 + self.redshift_source)

        if self.rest_frame_wavelengths is None:
            print(
                "please provide an array of extinction coefficients and an array of wavelengths"
            )
            return False

        extinction_interpolation = np.interp(
            rest_frame_wavelength,
            self.rest_frame_wavelengths,
            self.extinction_coefficients,
        )

        return extinction_interpolation

    def project_density_to_source_plane(self, inclination_angle):
        """Project the torus' density into a FluxProjection object.

        :param inclination_angle: inclination of the torus w.r.t. the observer in
            degrees
        :return: a 2 dimensional array representing the projected column density of the
            torus. Note that this is not representative of an emission or absorption, so
            a FluxProjection object is not returned.
        """

        projection = project_blr_to_source_plane(
            self.density_grid,
            np.zeros(np.shape(self.density_grid)),
            np.zeros(np.shape(self.density_grid)),
            inclination_angle,
            self.smbh_mass_exp,
            weighting_grid=self.density_grid,
            radial_resolution=self.radial_step,
            vertical_resolution=self.height_step,
        )[0]

        return projection

    def project_extinction_to_source_plane(
        self, inclination_angle, observer_frame_wavelength
    ):
        """Project the torus' modelled extinction to the source plane.

        :param inclination_angle: inclination of the torus w.r.t. the observer in
            degrees
        :param observer_frame_wavelength: observer frame wavelength in nm
        :return: FluxProjection object representing the amount of flux absorbed by the
            torus
        """

        extinction_strength = self.interpolate_to_extinction_at_wavelength(
            observer_frame_wavelength
        )
        projected_extinction_array = self.project_density_to_source_plane(
            inclination_angle
        )

        extinction_mask = projected_extinction_array > 0

        projected_extinction_array = np.nan_to_num(
            -2.5
            * np.log10(1 - np.exp(-projected_extinction_array / extinction_strength))
            * extinction_mask
        )

        projected_extinction = FluxProjection(
            projected_extinction_array,
            observer_frame_wavelength,
            self.smbh_mass_exp,
            self.redshift_source,
            np.max(self.radii_values),
            inclination_angle,
            OmM=self.OmM,
            H0=self.H0,
        )

        return projected_extinction



import numpy as np
import astropy.units as u
import astropy.constants as const
from speclite.filters import (
    load_filter,
    load_filters,
)
import speclite
from amoeba.Util.util import convolve_signal_with_transfer_function


def intrinsic_signal_propagation_pipeline_for_agn(
    AGN,
    intrinsic_light_curve=None,
    time_axis=None,
    observer_frame_wavelengths_in_nm=None,
    speclite_filter=None,
    return_components=False,
    **kwargs,
):
    """run the pipeline to generate the full AGN intrinsic signal

    :param AGN: The AGN object to propagate the signal through.
    :param intrinsic_light_curve: the signal to propogate though the AGN model. Must be in units of days and
        sampled daily if time_axis is not given.
    :param time_axis: the time stamps of the driving signal to be specified if the driving signal is not
        evenly sampled every day.
    :param observer_frame_wavelengths_in_nm: a wavelength or list of wavelengths in nm.
    :param speclite_filter: a speclite filter, list of speclite filters, or list of speclite filter names.
    :param return_components: a bool which allows the return of each component light curve in addition to
        the combined light curve.
    :param output_cadence: float/int representing the output cadence in days
    :return: a list with the length of wavelengths/filters of light curves represented by lists of time
        axes and amplitudes. If return_components, then the BLR light curves are returned as a dictionary
        with the key being the BLR index.
    """

    if observer_frame_wavelengths_in_nm is None and speclite_filter is None:
        print("please provide a range of wavelengths or a speclite filter to use")
        return False
    if observer_frame_wavelengths_in_nm is not None and speclite_filter is not None:
        print("only provide a range of wavelengths or a speclite filter to use")
        return False

    if speclite_filter is not None:
        if isinstance(speclite_filter, str):
            try:
                current_filters = [load_filter(speclite_filter)]
            except:
                current_filters = load_filters(speclite_filter)
        elif isinstance(speclite_filter, speclite.filters.FilterResponse):
            current_filters = [speclite_filter]
        elif isinstance(speclite_filter, list):
            successful_filters = []
            for item in speclite_filter:
                if isinstance(item, str):
                    try:
                        current_filter = load_filter(item)
                        successful_filters.append(current_filter)
                    except:
                        continue
                elif isinstance(item, speclite.filters.FilterResponse):
                    successful_filters.append(item)
                else:
                    print(f"{item} not recognized")
            if len(successful_filters) == 0:
                print("no filters loaded, no propagation required")
                return False
            current_filters = successful_filters

        mean_wavelengths = []
        wavelength_ranges = []

        for band in current_filters:
            mean_wavelengths.append(band.effective_wavelength.to(u.nm).value)
            min_wavelength = band.wavelength[np.argmax(band.response > 0.01)] / 10
            total_wavelengths = np.sum(band.response > 0.01)
            wavelength_ranges.append(
                [
                    int(min_wavelength),
                    int((min_wavelength + total_wavelengths / 10)),
                ]
            )

    else:
        if isinstance(observer_frame_wavelengths_in_nm, (int, float)):
            mean_wavelengths = [observer_frame_wavelengths_in_nm]
            wavelength_ranges = [
                [
                    observer_frame_wavelengths_in_nm - 20,
                    observer_frame_wavelengths_in_nm + 20,
                ]
            ]

        elif isinstance(observer_frame_wavelengths_in_nm, (list, np.ndarray)):
            mean_wavelengths = []
            wavelength_ranges = []

            for band in observer_frame_wavelengths_in_nm:
                if isinstance(band, (int, float)):
                    mean_wavelengths.append(band)
                    wavelength_ranges.append([band - 20, band + 20])
                elif isinstance(band, (list, np.ndarray)):
                    mean_wavelengths.append(np.mean(band))
                    wavelength_ranges.append([np.min(band), np.max(band)])

    if len(mean_wavelengths) == 0:
        print(
            "please provide a speclite filter, wavelength, wavelength range, or list containing \n previously mentioned types"
        )
        return False

    if "accretion_disk" not in AGN.components.keys():
        print(
            "please add an accretion disk model to this agn, other components require the variable continuum."
        )
        return False

    if intrinsic_light_curve is None:
        if AGN.intrinsic_light_curve is None:
            try:
                AGN.generate_intrinsic_signal(len(AGN.frequencies))
            except:
                print(
                    "please provide a psd and set of frequencies, or a driving light curve"
                )
                return False
        intrinsic_light_curve = AGN.intrinsic_light_curve.copy()
        time_axis = AGN.intrinsic_light_curve_time_axis.copy()

    reprocessed_signals = []
    for wavelength in mean_wavelengths:
        current_tf = AGN.components[
            "accretion_disk"
        ].construct_accretion_disk_transfer_function(wavelength)

        if "diffuse_continuum" in AGN.components.keys():
            current_dc_mean_lag_increase = AGN.components[
                "diffuse_continuum"
            ].get_diffuse_continuum_lag_contribution(wavelength)
            lag_increase = np.zeros(int(current_dc_mean_lag_increase))
            current_tf = np.concatenate((lag_increase, current_tf))

        t_ax, current_signal = convolve_signal_with_transfer_function(
            smbh_mass_exp=AGN.smbh_mass_exp,
            driving_signal=intrinsic_light_curve,
            initial_time_axis=time_axis,
            transfer_function=current_tf,
            redshift_source=0,
            desired_cadence_in_days=0.1,
        )
        reprocessed_signals.append([t_ax, current_signal])
    output_signals = reprocessed_signals.copy()

    blr_signals = {}

    if len(AGN.blr_indicies) > 0:

        for index in AGN.blr_indicies:
            blr_signals[str(index)] = []

            for jj, wavelength_range in enumerate(wavelength_ranges):
                weighting_factor, current_blr_tf = AGN.components[
                    "blr_" + str(index)
                ].calculate_blr_emission_line_transfer_function(
                    AGN.inclination_angle,
                    observed_wavelength_range_in_nm=wavelength_range,
                )

                t_ax, contaminated_signals = convolve_signal_with_transfer_function(
                    smbh_mass_exp=AGN.smbh_mass_exp,
                    driving_signal=reprocessed_signals[jj][1],
                    initial_time_axis=reprocessed_signals[jj][0],
                    transfer_function=current_blr_tf,
                    redshift_source=0,
                    desired_cadence_in_days=0.1,
                )

                contaminated_signals -= np.mean(contaminated_signals)
                if np.std(contaminated_signals) != 0:
                    contaminated_signals /= np.std(contaminated_signals)

                blr_signals[str(index)].append(
                    [t_ax, contaminated_signals, weighting_factor]
                )

    for jj, wavelength_range in enumerate(wavelength_ranges):

        current_weighting = 1
        original_mean = np.mean(reprocessed_signals[jj][1])
        original_std = np.std(reprocessed_signals[jj][1])

        current_signal = reprocessed_signals[jj][1] - original_mean
        if original_std != 0:
            current_signal /= original_std

        if len(AGN.blr_indicies) > 0:
            for index in AGN.blr_indicies:
                if isinstance(blr_signals[str(index)], list):
                    if not isinstance(
                        blr_signals[str(index)][jj], list
                    ):  # pragma: no cover
                        continue

                    current_weighting += blr_signals[str(index)][jj][-1]
                    current_blr_signal = blr_signals[str(index)][jj][1]
                    current_blr_signal -= np.mean(current_blr_signal)
                    if np.std(current_blr_signal) != 0:
                        current_blr_signal /= np.std(current_blr_signal)

                    current_signal += (
                        current_blr_signal * blr_signals[str(index)][jj][2]
                    )

        output_signals[jj] = [
            reprocessed_signals[jj][0] * (1 + AGN.redshift_source),
            current_signal,
        ]

    if return_components is True:
        return [reprocessed_signals, blr_signals, output_signals]
    return output_signals


def visualization_pipeline(
    AGN,
    inclination_angle=None,
    observer_frame_wavelengths_in_nm=None,
    speclite_filter=None,
    return_components=None,
    **kwargs,
):
    """This pipeline produces a flux projection object of the combined emission of each
    agn component. Note that depending on flux ratios between various components, one may easily
    dominate the flux distribution.

    :param AGN: The AGN object to project into the source plane
    :param inclination_angle: the inclination of the source w.r.t. the observer
    :param observer_frame_wavelengths_in_nm: a [list of] wavelength[s] to project the AGN into.
        Cannot be used with speclite_filter
    :param speclite_filter: a [list of] speclite filter[s] to project the AGN into. Cannot be
        used with observer_frame_wavelengths_in_nm.
    :param return_components: a boolean toggle to return each component as individual FluxProjection
        objects before adding them together.
    :return: A list of FluxProjection objects for each wavelength range or speclite filter given.
    """

    if inclination_angle is not None:
        update_output = AGN.update_inclination(inclination_angle)
        if update_output is False:
            return False

    if observer_frame_wavelengths_in_nm is None and speclite_filter is None:
        print("please provide a range of wavelengths or a speclite filter to use")
        return False
    if observer_frame_wavelengths_in_nm is not None and speclite_filter is not None:
        print("only provide a range of wavelengths or a speclite filter to use")
        return False

    if speclite_filter is not None:
        if isinstance(speclite_filter, str):
            try:
                current_filters = [load_filter(speclite_filter)]
            except:
                current_filters = load_filters(speclite_filter)
        elif isinstance(speclite_filter, speclite.filters.FilterResponse):
            current_filters = [speclite_filter]
        elif isinstance(speclite_filter, list):
            successful_filters = []
            for item in speclite_filter:
                if isinstance(item, str):
                    try:
                        current_filter = load_filter(item)
                        successful_filters.append(current_filter)
                    except:
                        continue
                elif isinstance(item, speclite.filters.FilterResponse):
                    successful_filters.append(item)
                else:
                    print(f"{item} not recognized")
            if len(successful_filters) == 0:
                print("no filters loaded, no propagation required")
                return False
            current_filters = successful_filters

        mean_wavelengths = []
        wavelength_ranges = []

        for band in current_filters:
            mean_wavelengths.append(band.effective_wavelength.to(u.nm).value)
            min_wavelength = band.wavelength[np.argmax(band.response > 0.01)] / 10
            total_wavelengths = np.sum(band.response > 0.01)
            wavelength_ranges.append(
                [
                    int(min_wavelength),
                    int((min_wavelength + total_wavelengths / 10)),
                ]
            )

    else:
        if isinstance(observer_frame_wavelengths_in_nm, (int, float)):
            mean_wavelengths = [observer_frame_wavelengths_in_nm]
            wavelength_ranges = [
                [
                    observer_frame_wavelengths_in_nm - 20,
                    observer_frame_wavelengths_in_nm + 20,
                ]
            ]

        elif isinstance(observer_frame_wavelengths_in_nm, (list, np.ndarray)):
            mean_wavelengths = []
            wavelength_ranges = []

            for band in observer_frame_wavelengths_in_nm:
                if isinstance(band, (int, float)):
                    mean_wavelengths.append(band)
                    wavelength_ranges.append([band - 20, band + 20])
                elif isinstance(band, (list, np.ndarray)):
                    mean_wavelengths.append(np.mean(band))
                    wavelength_ranges.append([np.min(band), np.max(band)])

    if len(mean_wavelengths) == 0:  # pragma: no cover
        print(
            "please provide a speclite filter, wavelength, wavelength range, or list containing \n previously mentioned types"
        )
        return False

    if len(AGN.components.keys()) == 0:
        print(
            "please add at least one coponent model to this agn. You cannot project nothing!"
        )
        return False

    list_of_wavelength_resolved_projections = []
    for jj, wavelength in enumerate(mean_wavelengths):
        list_of_projections = []
        if "accretion_disk" in AGN.components.keys():
            current_img = AGN.components[
                "accretion_disk"
            ].calculate_surface_intensity_map(wavelength)
            list_of_projections.append(current_img)

            reference_flux = current_img.total_flux

        if "diffuse_continuum" in AGN.components.keys():
            current_img = AGN.components[
                "diffuse_continuum"
            ].get_diffuse_continuum_emission(wavelength)
            list_of_projections.append(current_img)

            current_flux = current_img.total_flux

            if AGN.components["diffuse_continuum"].emissivity_etas is not None:
                emissivity = AGN.components[
                    "diffuse_continuum"
                ].interpolate_spectrum_to_wavelength(wavelength)
                current_img.flux_array *= emissivity * reference_flux / current_flux

        if len(AGN.blr_indicies) > 0:

            for kk, index in enumerate(AGN.blr_indicies):

                current_projection = AGN.components[
                    "blr_" + str(index)
                ].project_blr_intensity_over_velocity_range(
                    AGN.inclination_angle,
                    observed_wavelength_range_in_nm=[wavelength_ranges[jj]],
                )

                if current_projection.total_flux > 0:
                    current_projection.flux_array *= (
                        AGN.components["blr_" + str(index)].line_strength
                        * reference_flux
                        / current_projection.total_flux
                    )
                    current_projection.total_flux *= (
                        AGN.components["blr_" + str(index)].line_strength
                        * reference_flux
                        / current_projection.total_flux
                    )

                if kk == 0:
                    output_blr_projection = current_projection
                else:
                    output_blr_projection.add_flux_projection(current_projection)

            list_of_projections.append(output_blr_projection)
        list_of_wavelength_resolved_projections.append(list_of_projections)
    if return_components:
        return list_of_wavelength_resolved_projections

    output_projections = []
    for list_of_projections in list_of_wavelength_resolved_projections:
        output_projections.append(list_of_projections[0])
        for jj in range(len(list_of_projections) - 1):
            index = jj + 1
            output_projections[-1].add_flux_projection(list_of_projections[index])

    return output_projections



import numpy as np
import numpy.testing as npt
from speclite.filters import FilterResponse
from amoeba.Classes.accretion_disk import AccretionDisk
from amoeba.Classes.blr import BroadLineRegion
from amoeba.Classes.blr_streamline import Streamline
from amoeba.Classes.agn import Agn
from amoeba.Classes.diffuse_continuum import DiffuseContinuum
from amoeba.Classes.torus import Torus
from amoeba.Classes.flux_projection import FluxProjection
from amoeba.Util.util import (
    create_maps,
    convert_cartesian_to_polar,
)
from astropy import units as u
from astropy import constants as const


class TestAgn:

    def setup_method(self):

        self.init_smbh_mass_exp = 8.0
        self.init_redshift_source = 1.0
        self.init_inclination_angle = 0.0
        self.init_corona_height = 10
        self.init_number_grav_radii = 1000
        self.init_resolution = 1000
        self.init_spin = 0
        self.init_OmM = 0.3
        self.init_H0 = 70

        self.my_accretion_disk_kwargs = create_maps(
            self.init_smbh_mass_exp,
            self.init_redshift_source,
            self.init_number_grav_radii,
            self.init_inclination_angle,
            self.init_resolution,
            spin=self.init_spin,
            corona_height=self.init_corona_height,
        )

        self.accretion_disk = AccretionDisk(**self.my_accretion_disk_kwargs)

        x_vals = np.linspace(-2000, 2000, 100)
        X, Y = np.meshgrid(
            x_vals, x_vals / np.cos(self.init_inclination_angle * np.pi / 180)
        )
        R, Phi = convert_cartesian_to_polar(X, Y)

        self.radii_array = R
        self.phi_array = Phi
        cloud_density_radial_dependence = 0
        cloud_density_array = None
        r_in_in_gravitational_radii = 800
        r_out_in_gravitational_radii = 1000
        line_strength = 0.3
        frequencies = np.linspace(1 / 2000, 1 / 2, 100)
        power_law_density_dependence = -0.1
        name = "my diffuse continuum"

        self.my_dc_kwargs = {
            "smbh_mass_exp": self.init_smbh_mass_exp,
            "redshift_source": self.init_redshift_source,
            "inclination_angle": self.init_inclination_angle,
            "radii_array": self.radii_array,
            "phi_array": self.phi_array,
            "cloud_density_radial_dependence": cloud_density_radial_dependence,
            "cloud_density_array": cloud_density_array,
            "OmM": self.init_OmM,
            "H0": self.init_H0,
            "r_in_in_gravitational_radii": r_in_in_gravitational_radii,
            "r_out_in_gravitational_radii": r_out_in_gravitational_radii,
            "name": name,
            "line_strength": line_strength,
            "frequencies": frequencies,
            "power_law_density_dependence": power_law_density_dependence,
        }

        self.my_continuum = DiffuseContinuum(**self.my_dc_kwargs)

        self.init_launch_radius = 500  # Rg
        self.init_launch_theta = 0  # degrees
        self.init_max_height = 1000  # Rg
        self.init_height_step = 200
        self.init_line_strength = 0.3
        self.init_rest_frame_wavelength_in_nm = 600
        self.init_characteristic_distance = self.init_max_height // 5
        self.init_asymptotic_poloidal_velocity = 0.2
        self.init_poloidal_launch_velocity = 10**-5

        self.test_blr_streamline = Streamline(
            self.init_launch_radius,
            self.init_launch_theta,
            self.init_max_height,
            self.init_characteristic_distance,
            self.init_asymptotic_poloidal_velocity,
            poloidal_launch_velocity=self.init_poloidal_launch_velocity,
            height_step=self.init_height_step,
        )

        self.init_launch_theta_angled = 45
        self.test_blr_streamline_angled = Streamline(
            self.init_launch_radius,
            self.init_launch_theta_angled,
            self.init_max_height,
            self.init_characteristic_distance,
            self.init_asymptotic_poloidal_velocity,
            poloidal_launch_velocity=self.init_poloidal_launch_velocity,
            height_step=self.init_height_step,
        )

        self.test_torus_streamline_angled = Streamline(
            self.init_launch_radius * 10,
            self.init_launch_theta_angled,
            self.init_max_height,
            self.init_characteristic_distance,
            self.init_asymptotic_poloidal_velocity,
            poloidal_launch_velocity=self.init_poloidal_launch_velocity,
            height_step=self.init_height_step,
        )

        self.my_blr_kwargs = {
            "smbh_mass_exp": self.init_smbh_mass_exp,
            "max_height": self.init_max_height,
            "rest_frame_wavelength_in_nm": self.init_rest_frame_wavelength_in_nm,
            "redshift_source": self.init_redshift_source,
            "height_step": self.init_height_step,
            "line_strength": self.init_line_strength,
        }

        self.blr = BroadLineRegion(**self.my_blr_kwargs)

        self.blr.add_streamline_bounded_region(
            self.test_blr_streamline,
            self.test_blr_streamline_angled,
        )

        self.my_torus_kwargs = {
            "smbh_mass_exp": self.init_smbh_mass_exp,
            "max_height": self.init_max_height,
            "redshift_source": self.init_redshift_source,
            "height_step": self.init_height_step,
        }

        self.my_agn = Agn(
            agn_name="Wow, what an AGN.",
            **self.my_accretion_disk_kwargs,
        )

        self.my_blr_streamline_kwargs = {
            "InnerStreamline": self.test_blr_streamline,
            "OuterStreamline": self.test_blr_streamline_angled,
        }

        self.my_accretion_disk_kwargs["max_height"] = self.init_max_height
        self.my_accretion_disk_kwargs["power_spectrum"] = None

        self.my_populated_agn = Agn(
            agn_name="Amazing AGN",
            **self.my_accretion_disk_kwargs,
        )
        self.my_populated_agn.add_default_accretion_disk()
        self.my_populated_agn.add_diffuse_continuum(**self.my_dc_kwargs)
        self.my_populated_agn.add_blr(**self.my_blr_kwargs)
        self.my_populated_agn.add_streamline_bounded_region_to_blr(
            **self.my_blr_streamline_kwargs
        )

        self.my_populated_agn.add_torus(**self.my_torus_kwargs)

        self.extra_agn_kwargs = self.my_accretion_disk_kwargs.copy()
        del self.extra_agn_kwargs["H0"]
        del self.extra_agn_kwargs["OmM"]

        self.what_a_basic_agn = Agn(
            agn_name="Not so amazing AGN", **self.extra_agn_kwargs
        )

        min_kwargs = {
            "smbh_mass_exp": self.init_smbh_mass_exp,
            "redshift_source": self.init_redshift_source,
            "inclination_angle": self.init_inclination_angle,
        }
        self.no_kwarg_agn = Agn(agn_name="It's like a box", **min_kwargs)

    def test_initialization(self):

        assert self.my_agn.smbh_mass_exp == self.init_smbh_mass_exp
        assert self.my_agn.inclination_angle == self.init_inclination_angle
        assert self.my_agn.redshift_source == self.init_redshift_source
        assert self.my_agn.OmM == self.init_OmM
        assert self.my_agn.H0 == self.init_H0
        assert self.my_agn.smbh_mass == 10**self.init_smbh_mass_exp * const.M_sun.to(
            u.kg
        )
        assert self.my_agn.redshift_source == self.init_redshift_source
        assert self.my_agn.name == "Wow, what an AGN."
        assert self.my_agn.disk_is_updatable == True
        assert self.my_agn.blr_indicies == []
        assert self.my_agn.default_accretion_disk_kwargs is not None
        assert self.my_agn.generic_accretion_disk_kwargs is not None
        assert self.my_agn.blr_kwargs is not None
        assert self.my_agn.diffuse_continuum_kwargs is not None
        assert self.my_agn.intrinsic_signal_kwargs is not None
        assert self.my_agn.torus_kwargs is not None

    def test_add_default_accretion_disk(self):

        assert "accretion_disk" not in self.my_agn.components.keys()

        self.my_agn.add_default_accretion_disk()
        self.my_agn.add_default_accretion_disk(visc_temp_prof="NT")

        assert "accretion_disk" in self.my_agn.components.keys()
        assert self.my_agn.disk_is_updatable == True

        assert isinstance(self.my_agn.components["accretion_disk"], AccretionDisk)
        assert (
            self.my_agn.components["accretion_disk"].smbh_mass_exp
            == self.init_smbh_mass_exp
        )
        assert (
            self.my_agn.components["accretion_disk"].rg
            == self.my_agn.gravitational_radius
        )

    def test_add_generic_accretion_disk(self):

        agn_kwargs = self.my_accretion_disk_kwargs.copy()

        agn_kwargs["radii_array"] = self.radii_array
        agn_kwargs["phi_array"] = self.phi_array
        agn_kwargs["temp_array"] = 10 * self.radii_array
        agn_kwargs["g_array"] = np.ones(np.shape(self.radii_array))

        self.my_agn.add_generic_accretion_disk(**agn_kwargs)

        assert self.my_agn.disk_is_updatable == False
        assert "accretion_disk" in self.my_agn.components.keys()

        assert isinstance(self.my_agn.components["accretion_disk"], AccretionDisk)
        assert (
            self.my_agn.components["accretion_disk"].smbh_mass_exp
            == self.init_smbh_mass_exp
        )
        assert (
            self.my_agn.components["accretion_disk"].rg
            == self.my_agn.gravitational_radius
        )

    def test_add_blr(self):

        assert self.my_agn.add_blr(**self.my_blr_kwargs)

        assert "blr_0" in self.my_agn.components.keys()

        assert isinstance(self.my_agn.components["blr_0"], BroadLineRegion)

        self.my_blr_kwargs["line_strength"] = 2

        assert self.my_agn.add_blr(blr_index=1, **self.my_blr_kwargs)

        assert "blr_1" in self.my_agn.components.keys()
        assert 1 in self.my_agn.blr_indicies

        assert isinstance(self.my_agn.components["blr_1"], BroadLineRegion)

        assert (
            self.my_agn.components["blr_0"].line_strength
            < self.my_agn.components["blr_1"].line_strength
        )

    def test_add_streamline_bounded_region_to_blr(self):

        self.my_agn.add_blr(**self.my_blr_kwargs)
        self.my_agn.add_blr(blr_index=1, **self.my_blr_kwargs)

        self.one_blr_streamline = {"InnerStreamline": self.test_blr_streamline_angled}

        self.two_blr_streamlines = {
            "InnerStreamline": self.test_blr_streamline,
            "OuterStreamline": self.test_blr_streamline_angled,
        }

        # make sure that by adding only one streamline, we don't alter the blr
        assert False == self.my_agn.add_streamline_bounded_region_to_blr(
            0, **self.one_blr_streamline
        )

        assert np.size(self.my_agn.components["blr_" + str(0)].density_grid) == np.size(
            self.my_agn.components["blr_" + str(1)].density_grid
        )

        # show adding 2 streamlines physically changes the blr
        assert self.my_agn.add_streamline_bounded_region_to_blr(
            0, **self.two_blr_streamlines
        )

        assert np.size(self.my_agn.components["blr_" + str(0)].density_grid) > np.size(
            self.my_agn.components["blr_" + str(1)].density_grid
        )

        bad_blr_index = 33
        assert not self.my_agn.add_streamline_bounded_region_to_blr(
            blr_index=bad_blr_index, **self.two_blr_streamlines
        )

    def test_get_blr_density_axes(self):
        # we want to reliably get the axes for any type of input
        desired_index = 0
        list_of_axes = self.my_populated_agn.get_blr_density_axes(
            blr_index=desired_index
        )
        expected_shape = np.shape(
            self.my_populated_agn.components[
                "blr_" + str(desired_index)
            ].get_density_axis()
        )
        assert isinstance(list_of_axes, list)
        assert isinstance(list_of_axes[0], list)
        assert np.shape(list_of_axes[0]) == expected_shape
        all_axes = self.my_populated_agn.get_blr_density_axes()
        assert len(all_axes) == len(list_of_axes)
        assert isinstance(all_axes, list)
        assert isinstance(all_axes[0], list)
        assert np.shape(all_axes[0]) == expected_shape
        desired_index = "0"
        list_of_axes = self.my_populated_agn.get_blr_density_axes(
            blr_index=desired_index
        )
        assert isinstance(list_of_axes, list)
        assert isinstance(list_of_axes[0], list)
        assert np.shape(list_of_axes[0]) == expected_shape
        desired_index = [0]
        list_of_axes = self.my_populated_agn.get_blr_density_axes(
            blr_index=desired_index
        )
        assert isinstance(list_of_axes, list)
        assert isinstance(list_of_axes[0], list)
        assert np.shape(list_of_axes[0]) == expected_shape

        self.my_populated_agn.add_blr(blr_index=44.2, **self.my_blr_kwargs)

        all_axes = self.my_populated_agn.get_blr_density_axes()
        assert len(all_axes) > len(list_of_axes)

    def test_set_blr_efficiency_array(self):

        R = self.my_populated_agn.get_blr_density_axes()[0][0]
        test_efficiency = 1 / (1 + abs(R - 200))
        blr_index = 0
        assert np.shape(R) == np.shape(
            self.my_populated_agn.components["blr_0"].density_grid
        )
        assert not self.my_populated_agn.set_blr_efficiency_array()
        assert not self.my_populated_agn.set_blr_efficiency_array(
            efficiency_array=test_efficiency
        )
        assert not self.my_populated_agn.set_blr_efficiency_array(blr_index=blr_index)
        new_index = 22
        self.my_populated_agn.add_blr(blr_index=new_index)
        assert not self.my_populated_agn.set_blr_efficiency_array(blr_index=new_index)
        assert self.my_populated_agn.set_blr_efficiency_array(
            blr_index=blr_index,
            efficiency_array=test_efficiency,
        )
        assert not self.my_populated_agn.set_blr_efficiency_array(blr_index=blr_index)

    def test_add_torus(self):

        assert self.my_agn.add_torus(**self.my_torus_kwargs)

        assert isinstance(self.my_agn.components["torus"], Torus)

    def test_add_streamline_bounded_region_to_torus(self):

        self.my_agn.add_torus(**self.my_torus_kwargs)

        previous_density_grid_size = np.size(
            self.my_agn.components["torus"].density_grid
        )

        self.torus_streamline_kwargs = {"Streamline": self.test_torus_streamline_angled}

        assert self.my_agn.add_streamline_bounded_region_to_torus(
            **self.torus_streamline_kwargs
        )

        assert (
            np.size(self.my_agn.components["torus"].density_grid)
            > previous_density_grid_size
        )

        assert isinstance(self.my_agn.components["torus"], Torus)

        assert not self.my_agn.add_streamline_bounded_region_to_torus(
            **{"not_a_streamline": [1, 2, 3]}
        )

    def test_add_diffuse_continuum(self):

        additional_kwargs = {
            "cloud_density_radial_dependence": -0.5,
            "r_in_in_gravitational_radii": 400,
            "r_out_in_gravitational_radii": 5000,
        }

        self.my_agn.add_diffuse_continuum(**additional_kwargs)

        assert self.my_agn.components["diffuse_continuum"].responsivity_constant == 1
        assert (
            self.my_agn.components["diffuse_continuum"].rest_frame_wavelengths is None
        )
        assert self.my_agn.components["diffuse_continuum"].emissivity_etas is None

        additional_kwargs["responsivity_constant"] = 0.6

        assert self.my_agn.add_diffuse_continuum(**additional_kwargs)

        assert (
            self.my_agn.components["diffuse_continuum"].responsivity_constant
            == additional_kwargs["responsivity_constant"]
        )
        assert (
            self.my_agn.components["diffuse_continuum"].rest_frame_wavelengths is None
        )
        assert self.my_agn.components["diffuse_continuum"].emissivity_etas is None

        sample_wavelengths = [100, 500, 1000]
        sample_emissivities = [0.1, 0.4, 0.3]

        additional_kwargs["rest_frame_wavelengths"] = sample_wavelengths
        additional_kwargs["emissivity_etas"] = sample_emissivities

        assert self.my_agn.add_diffuse_continuum(**additional_kwargs)

        assert (
            self.my_agn.components["diffuse_continuum"].rest_frame_wavelengths
            is sample_wavelengths
        )

        assert (
            self.my_agn.components["diffuse_continuum"].emissivity_etas
            is sample_emissivities
        )

        # note the agn is prepared at redshift = 1
        test_emissivity_at_300_nm = self.my_agn.components[
            "diffuse_continuum"
        ].interpolate_spectrum_to_wavelength(300 * (1 + self.init_redshift_source))

        d_lam = (0.4 - 0.1) / (500 - 100)

        npt.assert_almost_equal(test_emissivity_at_300_nm, 0.1 + d_lam * 200)

        not_enough_kwargs = {"cloud_density_radial_dependence": -0.5}
        closer_kwargs = {
            "cloud_density_radial_dependence": -0.5,
            "radii_array": np.ones((100, 100)),
        }

        del self.my_agn.diffuse_continuum_kwargs["radii_array"]
        del self.my_agn.diffuse_continuum_kwargs["phi_array"]
        del self.my_agn.diffuse_continuum_kwargs["r_out_in_gravitational_radii"]

        with npt.assert_raises(AssertionError):
            self.my_agn.add_diffuse_continuum(**not_enough_kwargs)
        with npt.assert_raises(AssertionError):
            self.my_agn.add_diffuse_continuum(**closer_kwargs)

        extra_kwargs = self.my_dc_kwargs.copy()
        del extra_kwargs["radii_array"]
        assert self.my_agn.add_diffuse_continuum(**extra_kwargs)
        extra_kwargs = self.my_dc_kwargs.copy()
        del extra_kwargs["phi_array"]
        del self.my_agn.diffuse_continuum_kwargs["r_in_in_gravitational_radii"]
        assert "r_out_in_gravitational_radii" in extra_kwargs.keys()
        assert self.what_a_basic_agn.add_diffuse_continuum(**extra_kwargs)

        tiny_kwargs = {"r_out_in_gravitational_radii": 400}

        self.no_kwarg_agn.add_diffuse_continuum(**tiny_kwargs)

    def test_add_intrinsic_signal_parameters(self):

        length_of_light_curve = 1000
        cadence = 1

        # frequencies span up to nyquist frequency
        frequencies = np.linspace(
            1 / (2 * length_of_light_curve), 1 / (2 * cadence), length_of_light_curve
        )

        power_spectrum = (1 / frequencies) ** 2

        intrinsic_signal_kwargs = {
            "power_spectrum": power_spectrum,
            "frequencies": frequencies,
        }

        assert self.my_agn.add_intrinsic_signal_parameters(**intrinsic_signal_kwargs)

        assert self.my_agn.power_spectrum is power_spectrum
        assert self.my_agn.frequencies is frequencies
        assert self.my_agn.random_seed is None

        my_seed = 822

        intrinsic_signal_kwargs["random_seed"] = my_seed

        self.my_agn.add_intrinsic_signal_parameters(**intrinsic_signal_kwargs)

        assert self.my_agn.random_seed is my_seed

    def test_visualize_static_accretion_disk(self):

        wavelength_in_nm = 700

        # make sure it only works when you have an accretion disk
        with npt.assert_raises(KeyError):
            my_disk_flux_projection = self.my_agn.visualize_static_accretion_disk(
                wavelength_in_nm
            )

        self.my_agn.add_default_accretion_disk()

        my_disk_flux_projection = self.my_agn.visualize_static_accretion_disk(
            wavelength_in_nm
        )

        assert isinstance(my_disk_flux_projection, FluxProjection)
        assert my_disk_flux_projection.redshift_source == self.my_agn.redshift_source

        assert np.shape(my_disk_flux_projection.get_plotting_axes()) == np.shape(
            self.my_agn.components["accretion_disk"].get_plotting_axes()
        )

    def test_visualize_static_blr(self):

        with npt.assert_raises(KeyError):
            self.my_agn.visualize_static_blr()

        self.my_agn.add_blr(**self.my_blr_kwargs)
        self.blr.add_streamline_bounded_region(
            self.test_blr_streamline,
            self.test_blr_streamline_angled,
        )

        my_blr_projection = self.my_agn.visualize_static_blr()

        assert isinstance(my_blr_projection, FluxProjection)

        velocity_range = [0, 1]
        visualization_kwargs = {"velocity_range": velocity_range}

        my_approaching_blr_projection = self.my_agn.visualize_static_blr(
            **visualization_kwargs
        )

        assert isinstance(my_approaching_blr_projection, FluxProjection)
        assert np.sum(my_blr_projection.flux_array) != np.sum(
            my_approaching_blr_projection
        )

        # check superluminal velocities raise ValueError
        velocity_range = [1, 2]
        error_visualization_kwargs = {"velocity_range": velocity_range}

        with npt.assert_raises(ValueError):
            self.my_agn.visualize_static_blr(**error_visualization_kwargs)

    def test_visualize_torus_obscuration(self):

        with npt.assert_raises(KeyError):
            self.my_agn.visualize_torus_obscuration(50)

        self.my_agn.add_torus(**self.my_torus_kwargs)
        assert not self.my_agn.components[
            "torus"
        ].interpolate_to_extinction_at_wavelength(50)

        wavelengths = [100, 1000, 10000]
        absorption_strengths = [10**-10, 10**-10, 10**-5]

        self.my_torus_kwargs["rest_frame_wavelengths"] = wavelengths
        self.my_torus_kwargs["extinction_coefficients"] = absorption_strengths

        assert self.my_agn.add_torus(**self.my_torus_kwargs)

        self.torus_streamline_kwargs = {"Streamline": self.test_torus_streamline_angled}

        assert self.my_agn.add_streamline_bounded_region_to_torus(
            **self.torus_streamline_kwargs
        )

        wavelength = 1000

        my_torus_projection = self.my_agn.visualize_torus_obscuration(wavelength)

        assert isinstance(my_torus_projection, FluxProjection)

    def test_visualize_static_diffuse_continuum(self):

        sample_wavelengths = [100, 10000]
        sample_emissivities = [0.1, 1.0]

        additional_kwargs = {
            "cloud_density_radial_dependence": -0.5,
            "r_in_in_gravitational_radii": 400,
            "r_out_in_gravitational_radii": 5000,
        }

        additional_kwargs["rest_frame_wavelengths"] = sample_wavelengths
        additional_kwargs["emissivity_etas"] = sample_emissivities

        self.my_agn.add_diffuse_continuum(**additional_kwargs)

        observer_wavelength_in_nm_400 = 400
        observer_wavelength_in_nm_9001 = 9001

        projected_diffuse_continuum_400 = (
            self.my_agn.visualize_static_diffuse_continuum(
                observer_wavelength_in_nm_400
            )
        )

        projected_diffuse_continuum_9001 = (
            self.my_agn.visualize_static_diffuse_continuum(
                observer_wavelength_in_nm_9001
            )
        )

        assert isinstance(projected_diffuse_continuum_400, FluxProjection)

        assert np.shape(
            projected_diffuse_continuum_400.get_plotting_axes()
        ) == np.shape(projected_diffuse_continuum_9001.get_plotting_axes())

        assert (
            projected_diffuse_continuum_400.total_flux
            < projected_diffuse_continuum_9001.total_flux
        )

    def test_generate_intrinsic_signal(self):

        length_of_light_curve = 1000
        cadence = 1

        # frequencies span up to nyquist frequency
        frequencies = np.linspace(
            1 / (2 * length_of_light_curve), 1 / (2 * cadence), length_of_light_curve
        )

        power_spectrum = (1 / frequencies) ** 2

        intrinsic_signal_kwargs = {
            "power_spectrum": power_spectrum,
            "frequencies": frequencies,
        }

        assert self.my_agn.add_intrinsic_signal_parameters(**intrinsic_signal_kwargs)

        time_axis, my_signal = self.my_agn.generate_intrinsic_signal(
            length_of_light_curve
        )

        assert len(my_signal) == length_of_light_curve

        npt.assert_almost_equal(np.mean(my_signal), 0, 5)
        npt.assert_almost_equal(np.std(my_signal), 1, 5)

        time_axis, my_second_signal = self.my_agn.generate_intrinsic_signal(
            length_of_light_curve
        )

        assert np.sum((my_signal - my_second_signal) ** 2) != 0

        intrinsic_signal_kwargs["random_seed"] = 17
        self.my_agn.add_intrinsic_signal_parameters(**intrinsic_signal_kwargs)

        time_axis, my_seeded_signal = self.my_agn.generate_intrinsic_signal(
            length_of_light_curve
        )
        time_axis, my_second_seeded_signal = self.my_agn.generate_intrinsic_signal(
            length_of_light_curve, **intrinsic_signal_kwargs
        )

        assert np.sum((my_seeded_signal - my_second_seeded_signal) ** 2) == 0

        new_frequencies = np.linspace(
            1 / (length_of_light_curve), 1 / (cadence), 2 * length_of_light_curve
        )
        new_power_spectrum = (1 / new_frequencies) ** 2.5

        assert len(new_frequencies) != len(self.my_agn.frequencies)
        assert len(new_power_spectrum) != len(self.my_agn.power_spectrum)

        new_intrinsic_signal_kwargs = {
            "power_spectrum": new_power_spectrum,
            "frequencies": new_frequencies,
        }

        time_axis, my_third_signal = self.my_agn.generate_intrinsic_signal(
            2 * length_of_light_curve, **new_intrinsic_signal_kwargs
        )

        time_axis, my_fourth_signal = self.my_agn.generate_intrinsic_signal(
            2 * length_of_light_curve
        )

        assert np.sum((my_third_signal - my_fourth_signal) ** 2) == 0

        cadence = 0.2
        n_frequencies = int(length_of_light_curve / cadence)
        frequencies = np.linspace(
            1 / (2 * length_of_light_curve), 1 / (2 * cadence), n_frequencies
        )

        power_spectrum = (1 / frequencies) ** 2

        intrinsic_signal_kwargs = {
            "power_spectrum": power_spectrum,
            "frequencies": frequencies,
        }

        assert self.my_agn.add_intrinsic_signal_parameters(**intrinsic_signal_kwargs)

        time_axis, my_fifth_signal = self.my_agn.generate_intrinsic_signal(
            length_of_light_curve
        )

        assert len(my_fifth_signal) == length_of_light_curve / cadence

        # reset kwargs for testing purposes
        empty_dict_of_kwargs = {
            "power_spectrum": None,
            "frequencies": None,
            "random_seed": None,
        }

        self.my_agn.add_intrinsic_signal_parameters(**empty_dict_of_kwargs)

        assert not self.my_agn.generate_intrinsic_signal(length_of_light_curve)

    def test_update_smbh_mass_exponent(self):

        new_mass_exp = 3.3

        assert self.my_populated_agn.smbh_mass_exp != new_mass_exp

        for component in self.my_populated_agn.components.keys():
            assert (
                self.my_populated_agn.components[component].smbh_mass_exp
                != new_mass_exp
            )

        assert self.my_populated_agn.update_smbh_mass_exponent(new_mass_exp)

        assert self.my_populated_agn.smbh_mass_exp == new_mass_exp

        self.my_populated_agn.add_blr()

        for component in self.my_populated_agn.components.keys():
            assert (
                self.my_populated_agn.components[component].smbh_mass_exp
                == new_mass_exp
            )

        agn_kwargs = self.my_accretion_disk_kwargs.copy()

        agn_kwargs["radii_array"] = self.radii_array
        agn_kwargs["phi_array"] = self.phi_array
        agn_kwargs["temp_array"] = 10 * self.radii_array
        agn_kwargs["g_array"] = np.ones(np.shape(self.radii_array))

        self.my_agn.add_generic_accretion_disk(**agn_kwargs)

        assert self.my_agn.update_smbh_mass_exponent(new_mass_exp)

    def test_update_inclination(self):

        new_inclination = 42

        assert self.my_populated_agn.inclination_angle != new_inclination

        for component in self.my_populated_agn.components.keys():
            if (
                "inclination_angle"
                in vars(self.my_populated_agn.components[component]).keys()
            ):
                assert (
                    self.my_populated_agn.components[component].inclination_angle
                    != new_inclination
                )

        assert self.my_populated_agn.update_inclination(new_inclination)

        assert self.my_populated_agn.inclination_angle == new_inclination

        for component in self.my_populated_agn.components.keys():
            if (
                "inclination_angle"
                in vars(self.my_populated_agn.components[component]).keys()
            ):
                assert (
                    self.my_populated_agn.components[component].inclination_angle
                    == new_inclination
                )

        agn_kwargs = self.my_accretion_disk_kwargs.copy()

        agn_kwargs["radii_array"] = self.radii_array
        agn_kwargs["phi_array"] = self.phi_array
        agn_kwargs["temp_array"] = 10 * self.radii_array
        agn_kwargs["g_array"] = np.ones(np.shape(self.radii_array))

        self.my_agn.add_generic_accretion_disk(**agn_kwargs)

        assert not self.my_agn.update_inclination(new_inclination)

    def test_update_redshift(self):

        new_redshift = 200

        assert self.my_populated_agn.redshift_source != new_redshift

        for component in self.my_populated_agn.components.keys():
            assert self.my_populated_agn.redshift_source != new_redshift

        assert self.my_populated_agn.update_redshift(new_redshift)

        assert self.my_populated_agn.redshift_source == new_redshift

        for component in self.my_populated_agn.components.keys():
            assert self.my_populated_agn.redshift_source == new_redshift

    def test_update_h0(self):

        new_H0 = 1

        assert self.my_populated_agn.H0 != new_H0

        for component in self.my_populated_agn.components.keys():
            assert self.my_populated_agn.H0 != new_H0

        assert self.my_populated_agn.update_h0(new_H0)

        assert self.my_populated_agn.H0 == new_H0

        for component in self.my_populated_agn.components.keys():
            assert self.my_populated_agn.H0 == new_H0

    def test_update_omega_m(self):

        new_omega_m = 0.999

        assert self.my_populated_agn.OmM != new_omega_m

        for componet in self.my_populated_agn.components.keys():
            assert self.my_populated_agn.OmM != new_omega_m

        assert self.my_populated_agn.update_omega_m(new_omega_m)

        assert self.my_populated_agn.OmM == new_omega_m

        for componet in self.my_populated_agn.components.keys():
            assert self.my_populated_agn.OmM == new_omega_m

    def test_update_line_strength(self):

        new_line_strength = 40

        self.my_blr_kwargs["line_strength"] = 4
        self.my_populated_agn.add_blr(blr_index=88, **self.my_blr_kwargs)

        for index in self.my_populated_agn.blr_indicies:
            assert (
                self.my_populated_agn.components["blr_" + str(index)]
                != new_line_strength
            )

        for index in self.my_populated_agn.blr_indicies:
            assert self.my_populated_agn.update_line_strength(index, new_line_strength)

        for index in self.my_populated_agn.blr_indicies:
            assert (
                self.my_populated_agn.components["blr_" + str(index)].line_strength
                == new_line_strength
            )

    def test_calculate_accretion_disk_transfer_function(self):

        wavelength = 77

        assert not self.my_agn.calculate_accretion_disk_transfer_function(wavelength)

        my_disk_tf = self.my_populated_agn.calculate_accretion_disk_transfer_function(
            wavelength
        )

        assert isinstance(my_disk_tf, (list, np.ndarray))

        taus = np.linspace(0, len(my_disk_tf) - 1, len(my_disk_tf))
        mean = np.sum(my_disk_tf * taus) / np.sum(my_disk_tf)

        assert mean > 0

    def test_calculate_blr_transfer_function(self):

        wavelengths = [100, 2000]

        assert not self.my_agn.calculate_blr_transfer_function(wavelengths)

        my_blr_tfs = self.my_populated_agn.calculate_blr_transfer_function(wavelengths)

        weight, my_blr_tf = my_blr_tfs[0]

        assert isinstance(my_blr_tfs, (list, np.ndarray))
        assert isinstance(my_blr_tf, (list, np.ndarray))
        assert isinstance(weight, (float, int))

        taus = np.linspace(0, len(my_blr_tf) - 1, len(my_blr_tf))
        mean = np.sum(my_blr_tf * taus) / np.sum(my_blr_tf)
        assert mean > 0

    def test_calculate_diffuse_continuum_mean_time_lag(self):

        wavelength = 200 * np.pi

        assert not self.my_agn.calculate_diffuse_continuum_mean_time_lag(wavelength)

        lag_increase = self.my_populated_agn.calculate_diffuse_continuum_mean_time_lag(
            wavelength
        )

        assert lag_increase > 0

    def test_intrinsic_signal_propogation_pipeline(self):

        assert not self.my_populated_agn.intrinsic_signal_propagation_pipeline()
        assert not self.my_populated_agn.intrinsic_signal_propagation_pipeline(
            observer_frame_wavelengths_in_nm=500, speclite_filter="lsst2023-u"
        )

        assert not self.my_populated_agn.intrinsic_signal_propagation_pipeline(
            speclite_filter=["yellow"]
        )

        assert not self.my_populated_agn.intrinsic_signal_propagation_pipeline(
            observer_frame_wavelengths_in_nm=["cyan"]
        )

        assert not self.my_agn.intrinsic_signal_propagation_pipeline(
            speclite_filter="lsst2023-i"
        )

        assert not self.my_populated_agn.intrinsic_signal_propagation_pipeline(
            speclite_filter="lsst2023-*"
        )

        frequencies = np.linspace(1 / (2000), 1 / (2), 1000)

        power_spectrum = (1 / frequencies) ** 2

        intrinsic_signal_kwargs = {
            "power_spectrum": power_spectrum,
            "frequencies": frequencies,
        }

        assert self.my_populated_agn.add_intrinsic_signal_parameters(
            **intrinsic_signal_kwargs
        )

        all_my_signals = self.my_populated_agn.intrinsic_signal_propagation_pipeline(
            speclite_filter=["lsst2023-u", "lsst2023-z"], return_components=True
        )

        # I expect that this will return, for the u and z band, a list of the continuum, blr, and total signal
        # 3 components (continuum, blr, total)
        assert len(all_my_signals) == 3
        # each component has 2 bands (u, z)
        for jj in range(len(all_my_signals)):
            if isinstance(all_my_signals[jj], list):
                assert len(all_my_signals[jj]) == 2
            else:
                for key in all_my_signals[jj].keys():
                    # if we have a blr, we have instead 3 items in this list (time_axis, light_curve, weighting factor)
                    for band in range(len(all_my_signals[jj][key])):
                        assert len(all_my_signals[jj][key][band]) == 3
        # each component's band has 2 items (time_axis, light_curve), and each of these are equal in length
        # unless it's the blr, when there is a dictionary to allow for multiple blr components
        for jj in range(len(all_my_signals)):
            for kk in range(len(all_my_signals[jj])):
                if isinstance(all_my_signals[jj], list):
                    assert len(all_my_signals[jj][kk]) == 2
                    assert len(all_my_signals[jj][kk][0]) == len(
                        all_my_signals[jj][kk][1]
                    )
                else:
                    for key in all_my_signals[jj].keys():
                        assert len(all_my_signals[jj][key][0]) == len(
                            all_my_signals[jj][key][1]
                        )

        self.my_populated_agn.update_line_strength(blr_index=0, new_line_strength=0.3)

        # must also work for a set of wavelength ranges
        all_my_signals_2 = self.my_populated_agn.intrinsic_signal_propagation_pipeline(
            observer_frame_wavelengths_in_nm=[[100, 500], [500, 10000]],
            return_components=True,
        )

        assert len(all_my_signals_2) == 3
        # each component has 2 bands ([100-500], [500-10000])
        for jj in range(len(all_my_signals_2)):
            if isinstance(all_my_signals_2[jj], list):
                assert len(all_my_signals_2[jj]) == 2
            else:
                for key in all_my_signals_2[jj].keys():
                    for band in range(len(all_my_signals_2[jj][key])):
                        assert len(all_my_signals_2[jj][key][band]) == 3
        for jj in range(len(all_my_signals_2)):
            for kk in range(len(all_my_signals_2[jj])):
                if isinstance(all_my_signals_2[jj], list):
                    assert len(all_my_signals_2[jj][kk]) == 2
                    assert len(all_my_signals_2[jj][kk][0]) == len(
                        all_my_signals_2[jj][kk][1]
                    )
                else:
                    for key in all_my_signals_2[jj].keys():
                        assert len(all_my_signals_2[jj][key][0]) == len(
                            all_my_signals_2[jj][key][1]
                        )

        # must also work for a set of wavelengths
        all_my_signals_3 = self.my_populated_agn.intrinsic_signal_propagation_pipeline(
            observer_frame_wavelengths_in_nm=[100, 500, 10000, 750, 850],
            return_components=False,
        )

        assert len(all_my_signals_3) == 5
        # each component has 3 bands (100, 500, 10000, 750, 850)
        for jj in range(len(all_my_signals_3)):
            assert len(all_my_signals_3[jj]) == 2
        for jj in range(len(all_my_signals_3)):
            for kk in range(len(all_my_signals_3[jj])):
                assert len(all_my_signals_3[jj][0]) == len(all_my_signals_3[jj][1])

        # must also work for one wavelength range
        my_only_signal = self.my_populated_agn.intrinsic_signal_propagation_pipeline(
            observer_frame_wavelengths_in_nm=[[300, 700]]
        )

        assert len(my_only_signal) == 1
        assert len(my_only_signal[0]) == 2
        assert len(my_only_signal[0][0]) == len(my_only_signal[0][1])

        # and also one speclite filter
        my_other_only_signal = (
            self.my_populated_agn.intrinsic_signal_propagation_pipeline(
                speclite_filter="lsst2023-u"
            )
        )

        assert len(my_other_only_signal) == 1
        assert len(my_other_only_signal[0]) == 2
        assert len(my_other_only_signal[0][0]) == len(my_other_only_signal[0][1])

    def test_visualize_agn_pipeline(self):

        observer_frame_wavelengths_in_nm = [[1000, 1400]]
        # pick something very red which will have lower total flux
        no_blr_wavelengths = [[100000, 120000]]

        diffuse_continuum_wavelengths = [100, 10000]
        diffuse_continuum_spectrum = [1, 1]

        self.my_dc_kwargs["rest_frame_wavelengths"] = diffuse_continuum_wavelengths
        self.my_dc_kwargs["emissivity_etas"] = diffuse_continuum_spectrum

        self.my_populated_agn.add_diffuse_continuum(**self.my_dc_kwargs)

        components_with_blr = self.my_populated_agn.visualize_agn_pipeline(
            inclination_angle=None,
            observer_frame_wavelengths_in_nm=observer_frame_wavelengths_in_nm,
            return_components=True,
        )

        components_no_blr = self.my_populated_agn.visualize_agn_pipeline(
            inclination_angle=None,
            observer_frame_wavelengths_in_nm=no_blr_wavelengths,
            return_components=True,
        )

        projection_effect_with_blr = self.my_populated_agn.visualize_agn_pipeline(
            inclination_angle=None,
            observer_frame_wavelengths_in_nm=observer_frame_wavelengths_in_nm,
            return_components=False,
        )

        projection_effect_no_blr = self.my_populated_agn.visualize_agn_pipeline(
            inclination_angle=None,
            observer_frame_wavelengths_in_nm=no_blr_wavelengths,
            return_components=False,
        )

        assert isinstance(components_with_blr, list)
        assert isinstance(components_no_blr, list)
        assert isinstance(projection_effect_with_blr[0], FluxProjection)
        assert isinstance(projection_effect_no_blr[0], FluxProjection)
        assert isinstance(components_with_blr[0], list)
        assert isinstance(components_no_blr[0], list)
        assert isinstance(components_with_blr[0][0], FluxProjection)
        assert isinstance(components_no_blr[0][1], FluxProjection)

        assert len(components_with_blr[0]) == len(components_no_blr[0])
        assert np.shape(components_with_blr[0][-1].flux_array) > np.shape(
            components_no_blr[0][-1].flux_array
        )

        assert (
            projection_effect_with_blr[0].total_flux
            > projection_effect_no_blr[0].total_flux
        )

        # show that the pipeline works identically with and without returning components
        my_projection = self.my_populated_agn.visualize_agn_pipeline(
            inclination_angle=None,
            observer_frame_wavelengths_in_nm=observer_frame_wavelengths_in_nm,
            return_components=False,
        )
        assert my_projection[0].total_flux == projection_effect_with_blr[0].total_flux

        square_diff = np.sum(
            (projection_effect_with_blr[0].flux_array - my_projection[0].flux_array)
            ** 2
        )
        assert square_diff == 0

        # test that an agn without components will not return a projection
        my_null_projection = self.my_agn.visualize_agn_pipeline(
            observer_frame_wavelengths_in_nm=observer_frame_wavelengths_in_nm
        )
        assert my_null_projection is False

        # test that an agn with components but no wavelength will not return a projection
        my_null_projection = self.my_populated_agn.visualize_agn_pipeline()
        assert my_null_projection is False

        # test that we can project an agn in multiple bands with one call
        my_multitude_of_bands = ["lsst2023-u", "lsst2023-g", "lsst2023-z"]
        my_multi_components = self.my_populated_agn.visualize_agn_pipeline(
            speclite_filter=my_multitude_of_bands, return_components=True
        )
        my_multi_projection = self.my_populated_agn.visualize_agn_pipeline(
            speclite_filter=my_multitude_of_bands, return_components=False
        )
        assert len(my_multitude_of_bands) == len(my_multi_projection)
        # test we have the right number of components (no torus emission yet)
        assert len(self.my_populated_agn.components) == len(my_multi_components[0]) + 1
        # test each flux distribution is different
        assert my_multi_projection[0].total_flux != my_multi_projection[1].total_flux
        assert my_multi_projection[0].total_flux != my_multi_projection[2].total_flux

        # test that when inclination angle is updatable, you can update the agn object
        # directly with the projection pipeline
        new_inclination_angle = 40
        my_inclined_projection = self.my_populated_agn.visualize_agn_pipeline(
            inclination_angle=new_inclination_angle,
            speclite_filter=my_multitude_of_bands,
            return_components=False,
        )
        assert self.my_populated_agn.inclination_angle == new_inclination_angle

        for jj in range(len(my_multitude_of_bands)):
            assert (
                my_inclined_projection[jj].total_flux
                != my_multi_projection[jj].total_flux
            )
            assert (
                np.sum(
                    (
                        my_multi_projection[jj].flux_array
                        - my_inclined_projection[jj].flux_array
                    )
                    ** 2
                )
                != 0
            )



from astropy import units as u
from astropy import constants as const
import numpy as np
from amoeba.Classes.flux_projection import FluxProjection
from amoeba.Util.util import (
    calculate_luminosity_distance,
    calculate_gravitational_radius,
    planck_law,
    planck_law_derivative,
    calculate_time_lag_array,
    calculate_dt_dlx,
    construct_accretion_disk_transfer_function,
    generate_snapshots_of_radiation_pattern,
)


class AccretionDisk:

    def __init__(
        self,
        smbh_mass_exp=None,
        redshift_source=None,
        inclination_angle=None,
        corona_height=None,
        temp_array=None,
        phi_array=None,
        g_array=None,
        radii_array=None,
        height_array=None,
        albedo_array=None,
        spin=0,
        OmM=0.3,
        H0=70,
        r_out_in_gravitational_radii=None,
        name="",
        **kwargs
    ):
        """Object representing an accretion disk which is assumed to be optically thick.

        :param smbh_mass_exp: mass exponent of the sumpermassive black hole at the
            center of the disk expressed as log_10(M / M_sun). Typical ranges are 6-11
            for AGN.
        :param redshift_source: positive float representing the redshift of the AGN.
        :param inclination_angle: inclination of the accretion disk with respect to the
            observer, in degrees
        :param corona_height: height of the corona in the lamppost model in units of R_g
            = GM/c^2
        :param temp_array: a 2d representation of the effective temperature distribution
            of the accretion disk
        :param phi_array: a 2d representation of the azimuth angles on the accretion
            disk
        :param g_array: a 2d representation of the g-factors (relativistic correction
            factors) on the accretion disk. Requires general relativistic ray tracing.
        :param radii_array: a 2d representation of the radii on the accretion disk, in
            R_g = GM/c^2
        :param height_array: a 2d representation of the height of the accretion disk in
            units of R_g = GM/c^2
        :param spin: float representing the dimensionless spin component of the SMBH
            (often denoted a_{*}) which may range from -1 to 1. Negative values
            represent accretion disks anti-aligned with the spin of the black hole.
        :param OmM: Cosmological parameter representing the mass fraction of the
            universe
        :param H0: Hubble constant in units of km/s/Mpc
        :param r_out_in_gravitational_radii: maximum radius of the accretion disk, in
            R_g = GM/c^2
        :param name: Name space
        """

        self.name = name
        self.smbh_mass_exp = smbh_mass_exp
        self.mass = 10**smbh_mass_exp * const.M_sun.to(u.kg)
        self.redshift_source = redshift_source
        self.inclination_angle = inclination_angle
        self.spin = spin
        self.radii_array = radii_array
        self.height_array = height_array
        if albedo_array is None:
            albedo_array = np.zeros(np.shape(radii_array))
        elif isinstance(albedo_array, (int, float)):
            albedo_array = np.ones(np.shape(radii_array)) * albedo_array
        self.albedo_array = albedo_array
        if r_out_in_gravitational_radii is None:
            r_out_in_gravitational_radii = radii_array[0, np.size(radii_array, 0) // 2]
        self.r_out_in_gravitational_radii = r_out_in_gravitational_radii
        self.radial_mask = radii_array <= r_out_in_gravitational_radii
        self.temp_array = temp_array * self.radial_mask
        self.phi_array = phi_array * self.radial_mask
        self.g_array = g_array
        self.OmM = OmM
        self.H0 = H0
        self.lum_dist = calculate_luminosity_distance(
            self.redshift_source, OmM=self.OmM, H0=self.H0
        )
        self.rg = calculate_gravitational_radius(10**self.smbh_mass_exp)
        self.pixel_size = (
            self.rg
            * self.r_out_in_gravitational_radii
            * 2
            / np.size(self.temp_array, 0)
        )
        self.corona_height = corona_height

    def calculate_surface_intensity_map(
        self, observer_frame_wavelength_in_nm, return_wavelengths=False
    ):
        """Method to calculate the surface flux distribution at some wavelength.

        :param observer_frame_wavelength_in_nm: Wavelength in nanometers used to
            determine black body flux
        :param return_wavelengths: Bool used to return the map of wavelengths used at
            each pixel
        :return: A 2d array representing the surface flux density at desired wavelength.
            If returnwavelength is True, returns a tuple of 2d arrays.
        """

        if isinstance(observer_frame_wavelength_in_nm, u.Quantity):
            dummy = observer_frame_wavelength_in_nm.to(u.nm)
            observer_frame_wavelength_in_nm = dummy.value

        redshift_factor = 1 / (1 + self.redshift_source)
        total_redshift_factor = redshift_factor * self.g_array
        rest_frame_wavelength = total_redshift_factor * observer_frame_wavelength_in_nm

        output = (
            np.nan_to_num(
                planck_law(self.temp_array, rest_frame_wavelength)
                * pow(self.g_array, 4.0)
            )
            * self.pixel_size**2
        )
        if return_wavelengths == True:
            return output, rest_frame_wavelength
        flux_projection = FluxProjection(
            output,
            observer_frame_wavelength_in_nm,
            self.smbh_mass_exp,
            self.redshift_source,
            self.r_out_in_gravitational_radii,
            self.inclination_angle,
            OmM=self.OmM,
            H0=self.H0,
        )
        return flux_projection

    def calculate_db_dt_array(self, observer_frame_wavelength_in_nm):
        """Calculate the rate of change of surface flux density with respect to a
        fluctuation in temperature.

        :param observer_frame_wavelength_in_nm: Wavelength in nanometers in the observer
            frame
        :return: 2d array representing the partial derivative of the Planck function
            with respect to a small change in temperature at each pixel
        """

        if isinstance(observer_frame_wavelength_in_nm, u.Quantity):
            dummy = observer_frame_wavelength_in_nm.to(u.nm)
            observer_frame_wavelength_in_nm = dummy.value

        redshift_factor = 1 / (1 + self.redshift_source)
        total_redshift_factor = redshift_factor * self.g_array
        rest_frame_wavelength = total_redshift_factor * observer_frame_wavelength_in_nm

        output = (
            planck_law_derivative(self.temp_array, rest_frame_wavelength)
            * pow(self.g_array, 4.0)
            * self.pixel_size**2
        )
        return np.nan_to_num(output)

    def calculate_time_lag_array(
        self,
        corona_height=None,
        axis_offset_in_gravitational_radii=0,
        angle_offset_in_degrees=0,
    ):
        """Calculate the time delay between a point source and the accretion disk.

        :param corona_height: None or int / float. If None, use the corona height in the
            accretion disk. Otherwise, value represents height of the flare in units R_g
            = GM/c^2
        :param axis_offset_in_gravitational_radii: Radial offset of the flare with
            repsect to the axis of symmetry in units R_g = GM/c^2
        :param angle_offset_in_degrees: Angular rotation of the offset flare in degrees.
            Zero degrees represents a flare on the nearer side of the accretion disk,
            while 180 degrees represents the far side.
        :return: A 2d array of time delays representing the extra path length from the
            corona to the midplane of the accretion disk in units R_g / c.
        """

        if corona_height is None:
            corona_height = self.corona_height

        time_lag_array = calculate_time_lag_array(
            self.radii_array,
            self.phi_array,
            self.inclination_angle,
            corona_height=corona_height,
            axis_offset_in_gravitational_radii=axis_offset_in_gravitational_radii,
            angle_offset_in_degrees=angle_offset_in_degrees,
            height_array=self.height_array,
        )

        time_lag_array *= self.radial_mask

        return time_lag_array

    def calculate_dt_dlx_array(
        self,
        corona_height=None,
        axis_offset_in_gravitational_radii=0,
        angle_offset_in_degrees=0,
    ):
        """Calculate the change in temperature with respect to a change in the X-ray
        luminosity.

        :param corona_height: None or int / float. If None, the initialized corona
            height will be used. Otherwise, represents the height of the flare in units
            R_g = GM/c^2.
        :param axis_offset_in_gravitational_radii: Axis offset of the flaring event in
            units R_g.
        :param angle_offset_in_gravitational_radii: Degree rotation around the axis of
            symmetry of the flaring event. Zero degrees represents the flare nearer to
            the observer for inclined disks, while 180 degrees represnts the far side of
            the accretion disk.
        :return: a 2d array of values representing the partial derivative of accretion
            disk's temperature with respect to irradiating luminosity.
        """

        if corona_height is None:
            corona_height = self.corona_height

        dt_dlx_array = calculate_dt_dlx(
            self.temp_array,
            self.radii_array,
            self.phi_array,
            self.smbh_mass_exp,
            corona_height,
            axis_offset_in_gravitational_radii=axis_offset_in_gravitational_radii,
            angle_offset_in_degrees=angle_offset_in_degrees,
            height_array=self.height_array,
            albedo_array=self.albedo_array,
        )

        return dt_dlx_array

    def construct_accretion_disk_transfer_function(
        self,
        observer_frame_wavelength_in_nm,
        corona_height=None,
        axis_offset_in_gravitational_radii=0,
        angle_offset_in_degrees=0,
        return_response_array_and_lags=False,
    ):
        """Calculate the transfer function of the accretion disk within the lamppost
        model.

        :param observer_frame_wavelength_in_nm: Wavelength in nm with respect to the
            observer to calculate the transfer function at
        :param corona_height: None or int / float. If None, the initialized corona
            height will be used. Otherwise, represents the height of the flare in units
            R_g = GM/c^2
        :param axis_offset_in_gravitational_radii: Axis offset of the flaring event in
            units R_g = GM/c^2
        :param angle_offset_in_gravitational_radii: Degree rotation around the axis of
            symmetry of the flaring event. Zero degrees represents the flare nearer to
            the observer for inclined disks, while 180 degrees represnts the far side of
            the accretion disk
        :param return_response_array_and_lags: Boolean flag to return the accretion disk
            response and time delay arrays instead of the computed transfer function
        :return: 1d array representing the transfer function of the accretion disk with
            respect to some change in flux at a particular wavelength
        """
        if corona_height is None:
            corona_height = self.corona_height

        rest_frame_wavelength_in_nm = observer_frame_wavelength_in_nm / (
            1 + self.redshift_source
        )

        rest_frame_transfer_function = construct_accretion_disk_transfer_function(
            rest_frame_wavelength_in_nm,
            self.temp_array,
            self.radii_array,
            self.phi_array,
            self.g_array,
            self.inclination_angle,
            self.smbh_mass_exp,
            corona_height,
            axis_offset_in_gravitational_radii=axis_offset_in_gravitational_radii,
            angle_offset_in_degrees=angle_offset_in_degrees,
            height_array=self.height_array,
            albedo_array=self.albedo_array,
            return_response_array_and_lags=return_response_array_and_lags,
        )

        return rest_frame_transfer_function

    def generate_snapshots(
        self,
        observer_frame_wavelength_in_nm,
        time_stamps,
        driving_signal,
        driving_signal_fractional_strength,
        corona_height=None,
        axis_offset_in_gravitational_radii=0,
        angle_offset_in_degrees=0,
    ):
        """Method to generate snapshots of the accretion disk's surface brightness under
        the assumption that a driving signal is actively being reprocessed.

        :param observer_frame_wavelength_in_nm: Wavelength in nm in the observer's frame
            which we are observing the source at
        :param time_stamps: 1d array or list of times to pull the snapshots at
        :param driving_signal: 1d array or list representing the driving light curve
        :param driving_signal_fractional_strength: float representing how strong the
            reprocessed signal is with respect to the continuum emission
        :param corona_height: None or int/float. If None, the initialized corona height
            will be used. Otherwise, represents the height of the flare in units R_g =
            GM/c^2
        :param axis_offset_in_gravitational_radii: Axis offset of the flaring event in
            units R_g = GM/c^2
        :param angle_offset_in_gravitational_radii: Degree rotation around the axis of
            symmetry of the flaring event. Zero degrees represents the flare nearer to
            the observer for inclined disks, while 180 degrees represnts the far side of
            the accretion disk
        :return: a list of snapshots of the accretion disk at each time step. Note that
            this is an experimental method.
        """

        rest_frame_wavelength_in_nm = (
            observer_frame_wavelength_in_nm / (1 + self.redshift_source) / self.g_array
        )

        if corona_height is None:
            corona_height = self.corona_height

        radiation_patterns = generate_snapshots_of_radiation_pattern(
            rest_frame_wavelength_in_nm,
            time_stamps,
            self.temp_array,
            self.radii_array,
            self.phi_array,
            self.g_array,
            self.smbh_mass_exp,
            driving_signal,
            driving_signal_fractional_strength,
            corona_height,
            self.inclination_angle,
            axis_offset_in_gravitational_radii,
            angle_offset_in_degrees,
            self.height_array,
            self.albedo_array,
        )

        radiation_patterns_flux_projections = []

        for current_pattern in radiation_patterns:
            current_projection = FluxProjection(
                current_pattern,
                observer_frame_wavelength_in_nm,
                self.smbh_mass_exp,
                self.redshift_source,
                self.r_out_in_gravitational_radii,
                self.inclination_angle,
                OmM=self.OmM,
                H0=self.H0,
            )

            radiation_patterns_flux_projections.append(current_projection)

        return radiation_patterns_flux_projections

    def get_plotting_axes(self):
        """Method to get plotting axes for the accretion disk. Useful for plotting any
        of the generated arrays i.e. from calculate_time_lag_array,
        construct_accretion_disk_transfer_function( ...,

        return_response_array_and_lags=True ), self.temp_array, etc.

        :return: X, Y arrays to be used with plt.contourf()
        """

        x_ax = np.linspace(
            -self.r_out_in_gravitational_radii,
            self.r_out_in_gravitational_radii,
            np.size(self.radii_array, 0),
        )
        X, Y = np.meshgrid(x_ax, x_ax)

        return X, Y



import numpy as np
from amoeba.Util.util import (
    calculate_luminosity_distance,
    calculate_gravitational_radius,
    calculate_time_lag_array,
)
from amoeba.Classes.flux_projection import FluxProjection
import astropy.constants as const
import astropy.units as u


class DiffuseContinuum:

    def __init__(
        self,
        smbh_mass_exp=None,
        redshift_source=None,
        inclination_angle=None,
        radii_array=None,
        phi_array=None,
        cloud_density_radial_dependence=0,
        cloud_density_array=None,
        OmM=0.3,
        H0=70,
        r_in_in_gravitational_radii=None,
        r_out_in_gravitational_radii=None,
        name="",
        **kwargs
    ):
        """Object representing the diffuse continuum component of the AGN. The diffuse
        continuum model follows Korista+Goad 2019, where the increase in time lag
        is defined as:

        tau_lam(inci + DC) \approx tau_lam(DC) (1 - A)x / (1 - Ax)

        where:
        tau_lam(DC) is the time delay of the diffuse continuum and depends
            heavily on the cloud_density_array
        A is a constant on the range (0, 1)
        x is the fractional contribution of the DC to the total light

        These must be set using methods:
        A = self.set_responsivity_constant()
        x = self.set_emissivity()


        :param smbh_mass_exp: mass exponent of the sumpermassive black hole at the
            center of the disk expressed as log_10(M / M_sun). Typical ranges are 6-11
            for AGN.
        :param redshift_source: redshift of the diffuse continuum
        :param inclination_angle: inclination of the object w.r.t. the observer, in degrees
        :param radii_array: a 2d representation of the radii of the diffuse continuum, in
            gravitational radii R_g = GM/c^2
        :param phi_array: a 2d representation of the azimuths of the diffuse continuum
            in radians.
        :param cloud_density_radial_dependence: radial dependence of the clound density
            according to \rho \propto r^{alpha} where alpha is cloud_density_radial_dependence
        :param cloud_density_array: either an int, float, or 2d representation of the
            diffuse continuum cloud density.
        :param OmM: Cosmological parameter representing the mass fraction of the
            universe
        :param H0: Hubble constant in units of km/s/Mpc
        :param r_in_in_gravitational_radii: inner radius of the diffuse continuum in R_g.
            Note that this parameter dominates the time lag of the diffuse continuum.
        :param r_out_in_gravitational_radii: maximum radius of the diffuse continuum, in
            gravitational radii
        :param name: Name space
        """

        self.name = name
        self.smbh_mass_exp = smbh_mass_exp
        self.mass = 10**smbh_mass_exp * const.M_sun.to(u.kg)
        self.redshift_source = redshift_source
        self.inclination_angle = inclination_angle
        self.r_out_in_gravitational_radii = r_out_in_gravitational_radii
        self.r_in_in_gravitational_radii = r_in_in_gravitational_radii

        if r_out_in_gravitational_radii is not None:
            radial_mask = radii_array <= r_out_in_gravitational_radii
        else:
            radial_mask = np.ones(np.shape(radii_array))
            self.r_out_in_gravitational_radii = np.max(radii_array)

        if r_in_in_gravitational_radii is not None:
            radial_mask = radial_mask * (radii_array >= r_in_in_gravitational_radii)
            self.r_in_in_gravitational_radii = r_in_in_gravitational_radii
        else:
            self.r_in_in_gravitational_radii = 10

        self.radii_array = radii_array * radial_mask
        self.phi_array = phi_array
        self.radial_mask = radial_mask

        if cloud_density_array is not None:
            self.cloud_density_radial_dependence = None
            self.cloud_density_array = cloud_density_array * radial_mask
        elif isinstance(cloud_density_radial_dependence, (float, int)):
            self.cloud_density_array = (
                np.nan_to_num(self.radii_array**cloud_density_radial_dependence)
                * self.radial_mask
            )
            self.cloud_density_radial_dependence = cloud_density_radial_dependence
        else:
            raise ValueError(
                "Cloud density array and cloud density radial dependence cannot both be None type"
            )

        self.OmM = OmM
        self.H0 = H0
        self.lum_dist = calculate_luminosity_distance(
            self.redshift_source, OmM=self.OmM, H0=self.H0
        )
        self.rg = calculate_gravitational_radius(10**self.smbh_mass_exp)
        self.pixel_size = (
            self.rg
            * self.r_out_in_gravitational_radii
            * 2
            / np.size(self.radii_array, 0)
        )
        self.kwargs = kwargs

        if "emissivity_etas" in self.kwargs and "rest_frame_wavelengths" in self.kwargs:
            self.set_emissivity(
                rest_frame_wavelengths=self.kwargs["rest_frame_wavelengths"],
                emissivity_etas=self.kwargs["emissivity_etas"],
            )
        else:
            self.set_emissivity()

        if "responsivity_constant" in self.kwargs:
            self.set_responsivity_constant(
                responsivity_constant=self.kwargs["responsivity_constant"]
            )
        else:
            self.set_responsivity_constant()

    def set_emissivity(self, rest_frame_wavelengths=None, emissivity_etas=None):
        """Define the diffuse continuum's emission spectrum according to a rest frame
        wavelength spectrum. Required for estimating the increase in time lags.
        Determines the value of "x" in the equation in the init function.

        :param rest_frame_wavelengths: list representing the rest frame wavelengths in
            nanometers. Should cover the range of wavelengths expected to be observed.
        :param emissivity_etas: list representing the emissivities of the diffuse
            continuum clouds at each wavelength in rest_frame_wavelengths. Must be the
            same length as rest_frame_wavelengths.
        :return: True if successful
        """

        self.rest_frame_wavelengths = rest_frame_wavelengths
        self.emissivity_etas = emissivity_etas

        return True

    def set_responsivity_constant(self, responsivity_constant=1):
        """Define the responsivity constant of the BLR according to Korista+Goad 2019.
        Sets the value of "A" in the equation in the init function.

        :param responsivity_constant: define A, the percentage of flux that comes from
            the diffuse continuum w.r.t. the total flux. Must be between 0 and 1.
        :return: True if successful
        """

        assert responsivity_constant >= 0
        assert responsivity_constant <= 1

        self.responsivity_constant = responsivity_constant

        return True

    def interpolate_spectrum_to_wavelength(self, observer_frame_wavelength_in_nm):
        """Interpolates known spectra to a particular wavelength via linear
        interpolation.

        :param observer_frame_wavelength_in_nm: observer frame wavelength in nm
        :return: emissivity at observer frame wavenegth
        """

        if self.rest_frame_wavelengths is None:
            print("please initialize the diffuse continuum spectrum")
            return False

        rest_frame_wavelength_in_nm = observer_frame_wavelength_in_nm / (
            1 + self.redshift_source
        )
        emissivity_interpolation = np.interp(
            rest_frame_wavelength_in_nm,
            self.rest_frame_wavelengths,
            self.emissivity_etas,
        )

        return emissivity_interpolation

    def get_diffuse_continuum_emission(
        self, observer_frame_wavelength_in_nm, incident_continuum_weighting=1
    ):
        """Produce a FluxProjection object of the diffuse continuum under the assumption
        that there is no inclination dependence. Primarily used for the AGN projection
        pipeline. Note that there is a geometric weighting of r^{-2} from the incident
        light.

        :param observer_frame_wavelength_in_nm: wavelength the diffuse continuum is
            observed at in nanometers
        :param incident_continuum_weighting: optional weighting factor to scale the flux
            array by
        :return: FluxProjection object containing metadata of the diffuse continuum and
            the emission array
        """
        rest_frame_wavelength_in_nm = observer_frame_wavelength_in_nm / (
            1 + self.redshift_source
        )
        emission_array = self.cloud_density_array * np.nan_to_num(
            self.radii_array ** (-2)
        )
        emission_array /= np.sum(emission_array)
        emission_array *= (
            self.interpolate_spectrum_to_wavelength(rest_frame_wavelength_in_nm)
            * incident_continuum_weighting
        )

        projection = FluxProjection(
            emission_array * self.pixel_size ** (-2),
            observer_frame_wavelength_in_nm,
            self.smbh_mass_exp,
            self.redshift_source,
            self.r_out_in_gravitational_radii,
            self.inclination_angle,
            OmM=self.OmM,
            H0=self.H0,
        )

        return projection

    def get_diffuse_continuum_mean_lag(self, observer_frame_wavelength_in_nm):
        """Calculate the diffuse continuum's mean time lag at a given wavelength in the
        observer's frame. The integration constant is assumed to enforce tau_min = R_in
        / c.

        :param observer_frame_wavelength_in_nm: observer frame wavelength in nanometers
        :return: float representing the mean increase in time delay in units R_g / c
        """

        if self.cloud_density_radial_dependence is not None:

            integration_constant = self.r_in_in_gravitational_radii

            if self.cloud_density_radial_dependence < 0:
                integration_multiplicative_constant = (2 * np.pi) / (
                    self.cloud_density_radial_dependence
                )

                return (
                    integration_multiplicative_constant
                    * (
                        self.r_out_in_gravitational_radii
                        ** (self.cloud_density_radial_dependence)
                        - self.r_in_in_gravitational_radii
                        ** (self.cloud_density_radial_dependence)
                    )
                    + integration_constant
                )

            elif self.cloud_density_radial_dependence == 0:
                integration_multiplicative_constant = 2 * np.pi
                return (
                    integration_multiplicative_constant
                    * np.log(
                        self.r_out_in_gravitational_radii
                        / self.r_in_in_gravitational_radii
                    )
                    + integration_constant
                )

            else:
                integration_multiplicative_constant = (2 * np.pi) / (
                    self.cloud_density_radial_dependence
                )
                return (
                    integration_multiplicative_constant
                    * (
                        self.r_out_in_gravitational_radii
                        ** (self.cloud_density_radial_dependence)
                        - self.r_in_in_gravitational_radii
                        ** (self.cloud_density_radial_dependence)
                    )
                    + integration_constant
                )

        else:
            time_delays = calculate_time_lag_array(
                self.radii_array, self.phi_array, 0, 0
            )
            emissivity = self.get_diffuse_continuum_emission(
                observer_frame_wavelength_in_nm
            ).flux_array

            diffuse_continuum_transfer_function = np.histogram(
                time_delays,
                range=(0, np.max(time_delays)),
                bins=int(np.max(time_delays)),
                weights=np.nan_to_num(emissivity),
                density=True,
            )[0]

            diffuse_continuum_transfer_function[0] = 0

            diffuse_continuum_transfer_function /= np.sum(
                diffuse_continuum_transfer_function
            )

            mean_delay = np.sum(
                np.linspace(0, np.max(time_delays) - 1, int(np.max(time_delays)))
                * diffuse_continuum_transfer_function
            )

            return mean_delay

    def get_diffuse_continuum_lag_contribution(
        self,
        observer_frame_wavelength_in_nm,
    ):
        """Estimate the mean increase in time lag for an accretion disk due to the
        diffuse continuum, which must be added by the mean lag of the continuum. Solves
        the equation from Korista+Goad 2019 in the init function.

        :param observer_frame_wavelength_in_nm: observer frame wavelength in nanometers
        :return: increase in mean time lag at an observer frame wavelength in R_g / c
        """

        tau_diffuse = self.get_diffuse_continuum_mean_lag(
            observer_frame_wavelength_in_nm
        )
        emissivity = self.interpolate_spectrum_to_wavelength(
            observer_frame_wavelength_in_nm
        )
        const_a = self.responsivity_constant

        tau_dc_vs_continuum = (
            tau_diffuse * (1 - const_a) * emissivity / (1 - const_a * emissivity)
        )

        return tau_dc_vs_continuum



import numpy as np
from astropy import units as u
from astropy import constants as const
from amoeba.Util.util import (
    project_blr_to_source_plane,
    calculate_blr_transfer_function,
    determine_emission_line_velocities,
    calculate_keplerian_velocity,
    calculate_gravitational_radius,
    convert_speclite_filter_to_wavelength_range,
)
from amoeba.Classes.flux_projection import FluxProjection


class BroadLineRegion:

    def __init__(
        self,
        smbh_mass_exp,
        max_height,
        rest_frame_wavelength_in_nm,
        redshift_source,
        radial_step=10,
        height_step=10,
        max_radius=0,
        OmM=0.3,
        H0=70,
        line_strength=1,
        **kwargs
    ):
        """Generate a broad line region object. This object contains the R-Z
        distributions of clouds that lead to emission. Note that once initialized, you
        must add information to it using the Streamline objects.

        :param smbh_mass_exp: solution to log_10(m_smbh / m_sun)
        :param max_height: max Z coordinate of the BLR in units of R_g
        :param rest_frame_wavelength_in_nm: the natural wavelength of emission in nm.
            This will get Doppler shifted and redshifted to the observer's reference
            frame.
        :param redshift_source: redshift of the BLR object.
        :param radial_step: resolution of the BLR in the R dimension, in R_g
        :param height_step: resolution of the BLR in the Z direction, in R_g
        :param max_radius: maximum radius of the BLR in R_g
        :param OmM: mass component of the energy budget of the universe
        :param H0: Hubble constant in units of km/s/Mpc
        :param line_strength: float/int representing how strong the total emission line
            is w.r.t. the continuum emission.
        """

        self.smbh_mass_exp = smbh_mass_exp
        self.rg = calculate_gravitational_radius(10**self.smbh_mass_exp)
        self.rest_frame_wavelength_in_nm = rest_frame_wavelength_in_nm
        self.redshift_source = redshift_source
        self.OmM = OmM
        self.H0 = H0
        self.line_strength = line_strength
        self.max_height = max_height
        self.radial_step = radial_step
        self.height_step = height_step
        self.max_radius = max_radius
        self.density_grid = np.zeros(
            (max_radius // radial_step + 1, max_height // height_step + 1)
        )
        self.emission_efficiency_array = None
        self.current_calculated_inclination = None
        self.current_total_emission = None
        self.z_velocity_grid = np.zeros(np.shape(self.density_grid))
        self.r_velocity_grid = np.zeros(np.shape(self.density_grid))
        self.radii_values = np.linspace(0, max_radius, max_radius // radial_step + 1)
        self.height_values = np.linspace(0, max_height, max_height // height_step + 1)

        self.blr_array_shape = np.shape(self.density_grid)

        self.mass = 10 ** (self.smbh_mass_exp) * const.M_sun.to(u.kg)

    def add_streamline_bounded_region(
        self, InnerStreamline, OuterStreamline, density_initial_weighting=1
    ):
        """Add a region to the BLR which represents the line emitting region.

        :param InnerStreamline: Streamline object representative of the inner boundary
            conditions.
        :param OuterStreamline: Streamline object representative of the outer boundary
            conditions.
        :param density_initial_weighting: weighting factor for the density grid. Useful
            when defining multiple regions in the same BLR object. All weighting is
            relative.
        :return: True if successful
        """

        assert InnerStreamline.height_step == OuterStreamline.height_step
        assert InnerStreamline.max_height == OuterStreamline.max_height
        assert InnerStreamline.height_step == self.height_step
        assert InnerStreamline.max_height == self.max_height
        assert density_initial_weighting > 0

        if (
            np.max(
                [
                    np.max(InnerStreamline.radii_values),
                    np.max(OuterStreamline.radii_values),
                ]
            )
            > self.max_radius
        ):
            previous_maximum = self.max_radius // self.radial_step + 1
            self.max_radius = int(
                np.max(
                    [
                        np.max(InnerStreamline.radii_values),
                        np.max(OuterStreamline.radii_values),
                    ]
                )
            )
            dummygrid = np.zeros(
                (
                    self.max_radius // self.radial_step + 1,
                    self.max_height // self.height_step + 1,
                )
            )
            dummygrid[:previous_maximum, :] = self.density_grid
            self.density_grid = dummygrid
            dummygrid = np.zeros(np.shape(dummygrid))
            dummygrid[:previous_maximum, :] = self.z_velocity_grid
            self.z_velocity_grid = dummygrid
            dummygrid = np.zeros(np.shape(dummygrid))
            dummygrid[:previous_maximum, :] = self.r_velocity_grid
            self.r_velocity_grid = dummygrid

            self.radii_values = np.linspace(
                0, self.max_radius, self.max_radius // self.radial_step + 1
            )

        for hh in range(np.size(self.height_values)):
            low_mask = self.radii_values >= min(
                InnerStreamline.radii_values[hh], OuterStreamline.radii_values[hh]
            )
            high_mask = self.radii_values <= max(
                InnerStreamline.radii_values[hh], OuterStreamline.radii_values[hh]
            )
            mask = np.logical_and(low_mask, high_mask) + 0
            if hh == 0:
                norm = sum(mask)
            self.density_grid[np.argmax(mask) : np.argmax(mask) + sum(mask), hh] *= 0
            self.r_velocity_grid[np.argmax(mask) : np.argmax(mask) + sum(mask), hh] *= 0
            self.z_velocity_grid[np.argmax(mask) : np.argmax(mask) + sum(mask), hh] *= 0

            kkspace = np.linspace(0, sum(mask), sum(mask))
            self.z_velocity_grid[
                np.argmax(mask) : np.argmax(mask) + sum(mask), hh
            ] = InnerStreamline.poloidal_velocity[hh] * np.cos(
                InnerStreamline.launch_theta
            ) + (
                kkspace / sum(mask)
            ) * (
                OuterStreamline.poloidal_velocity[hh]
                * np.cos(OuterStreamline.launch_theta)
                - InnerStreamline.poloidal_velocity[hh]
                * np.cos(InnerStreamline.launch_theta)
            )
            self.r_velocity_grid[
                np.argmax(mask) : np.argmax(mask) + sum(mask), hh
            ] = InnerStreamline.poloidal_velocity[hh] * np.sin(
                InnerStreamline.launch_theta
            ) + (
                kkspace / sum(mask)
            ) * (
                OuterStreamline.poloidal_velocity[hh]
                * np.sin(OuterStreamline.launch_theta)
                - InnerStreamline.poloidal_velocity[hh]
                * np.sin(InnerStreamline.launch_theta)
            )

            del_pol_vels_on_vels = InnerStreamline.dpol_vel_dz_on_vel[hh] + (
                kkspace / sum(mask)
            ) * (
                OuterStreamline.dpol_vel_dz_on_vel[hh]
                - InnerStreamline.dpol_vel_dz_on_vel[hh]
            )
            self.density_grid[np.argmax(mask) : np.argmax(mask) + sum(mask), hh] += (
                density_initial_weighting
                * del_pol_vels_on_vels
                / self.radii_values[np.argmax(mask) : np.argmax(mask) + sum(mask)]
            )
        self.blr_array_shape = np.shape(self.density_grid)

        if self.emission_efficiency_array is not None:
            old_efficiency_array = self.emission_efficiency_array
            new_efficiency_array = np.zeros(self.blr_array_shape)
            new_efficiency_array[
                : np.size(old_efficiency_array, 0), : np.size(old_efficiency_array, 1)
            ] = old_efficiency_array

            self.set_emission_efficiency_array(new_efficiency_array)

        return True

    def project_blr_density_to_source_plane(self, inclination_angle):
        """Project the total density of the BLR into the source plane. Creates a
        FluxProjection object where the flux_array is representative of the density.

        :param inclination_angle: inclination of the source in degrees
        :return: FluxProjection containing metadata from the BLR and the emission array
        """

        projection, projected_number_gravitational_radii = project_blr_to_source_plane(
            self.density_grid,
            self.z_velocity_grid,
            self.r_velocity_grid,
            inclination_angle,
            self.smbh_mass_exp,
            weighting_grid=self.density_grid,
            radial_resolution=self.radial_step,
            vertical_resolution=self.height_step,
        )

        projection_object = FluxProjection(
            projection,
            self.rest_frame_wavelength_in_nm,
            self.smbh_mass_exp,
            self.redshift_source,
            projected_number_gravitational_radii,
            inclination_angle,
            OmM=self.OmM,
            H0=self.H0,
        )

        return projection

    def project_blr_total_intensity(
        self,
        inclination_angle,
        emission_efficiency_array=None,
    ):
        """Project the BLR into the source plane. Similar to
        project_blr_density_to_source_plane, but can be weighted by the emission
        efficiency which is an array of equal size to the BLR's density array.

        :param inclination_angle: inclination which we view the source at, in degrees
        :param emission_efficiency_array: a 2d grid of weighting factors in the R-Z
            plane.
        :return: FluxProjection containing metadata from the BLR and the emission array
        """

        self.set_emission_efficiency_array(emission_efficiency_array)

        flux_map, _ = project_blr_to_source_plane(
            self.density_grid,
            self.z_velocity_grid,
            self.r_velocity_grid,
            inclination_angle,
            self.smbh_mass_exp,
            weighting_grid=self.emission_efficiency_array,
            radial_resolution=self.radial_step,
            vertical_resolution=self.height_step,
        )

        flux_map *= self.line_strength

        max_radius = (
            self.max_height * np.tan(inclination_angle * np.pi / 180) + self.max_radius
        )

        flux_projection = FluxProjection(
            flux_map,
            np.array([0, np.inf]),
            self.smbh_mass_exp,
            self.redshift_source,
            max_radius,
            inclination_angle,
            OmM=self.OmM,
            H0=self.H0,
        )

        self.current_calculated_inclination = inclination_angle
        self.current_total_emission = flux_projection.total_flux

        return flux_projection

    def project_blr_intensity_over_velocity_range(
        self,
        inclination_angle,
        velocity_range=None,
        observed_wavelength_range_in_nm=None,
        speclite_filter=None,
        emission_efficiency_array=None,
    ):
        """Project a portion of the BLR to the source plane by determining the Doppler
        shifting at each coordinate in R-Z-phi space.

        :param inclination_angle: orientation of the source with respect to the observer
            in degrees
        :param velocity_range: a list representing the min and max velocities to choose.
            Cannot be used with observed_wavelength_range_in_nm or speclite_filter. Note
            that positive velocity is toward the observer.
        :param observed_wavelength_range_in_nm: a list representing the min and max
            wavelengths in nanometers which are observed. Cannot be used with
            velocity:range or speclite_filter.
        :param speclite_filter: Speclite filter object or string representing a loadable
            Speclite filter. Cannot be used with velocity_range or
            observed_wavelength_range_in_nm.
        :param emission_efficiency_array: a 2d grid of weighting factors in the R-Z
            plane
        :return: FluxProjection containing metadata from the BLR and the velocity
            selected emission array
        """

        if velocity_range is not None and observed_wavelength_range_in_nm is not None:
            print("Please only provide the velocities or wavelengths. Not both!")
            return False
        if (
            velocity_range is None
            and observed_wavelength_range_in_nm is None
            and speclite_filter is None
        ):
            print("Please provide the velocities or wavelengths.")
            velocity_range = [-1, 1]
        if speclite_filter is not None:
            observed_wavelength_range_in_nm = (
                convert_speclite_filter_to_wavelength_range(
                    speclite_filter,
                )
            )
        if observed_wavelength_range_in_nm is not None:
            velocity_range = determine_emission_line_velocities(
                self.rest_frame_wavelength_in_nm,
                np.min(observed_wavelength_range_in_nm),
                np.max(observed_wavelength_range_in_nm),
                self.redshift_source,
            )

        self.set_emission_efficiency_array(emission_efficiency_array)

        obs_plane_wavelength_in_nm = self.rest_frame_wavelength_in_nm * (
            1 + self.redshift_source
        )
        min_obs_plane_wavelength_in_nm = (
            obs_plane_wavelength_in_nm
            * ((1 - np.max(velocity_range)) / (1 + np.max(velocity_range))) ** 0.5
        )
        max_obs_plane_wavelength_in_nm = (
            obs_plane_wavelength_in_nm
            * ((1 - np.min(velocity_range)) / (1 + np.min(velocity_range))) ** 0.5
        )

        expected_broadening = self.estimate_doppler_broadening(inclination_angle)

        min_expected_wavelength_in_nm = (
            obs_plane_wavelength_in_nm
            * ((1 - np.max(expected_broadening)) / (1 + np.max(expected_broadening)))
            ** 0.5
        )
        max_expected_wavelength_in_nm = (
            obs_plane_wavelength_in_nm
            * ((1 - np.min(expected_broadening)) / (1 + np.min(expected_broadening)))
            ** 0.5
        )

        if min_expected_wavelength_in_nm > max_obs_plane_wavelength_in_nm:
            return FluxProjection(
                np.zeros((100, 100)),
                [min_obs_plane_wavelength_in_nm, max_obs_plane_wavelength_in_nm],
                self.smbh_mass_exp,
                self.redshift_source,
                100,
                inclination_angle,
                OmM=self.OmM,
                H0=self.H0,
            )
        if max_expected_wavelength_in_nm < min_obs_plane_wavelength_in_nm:
            return FluxProjection(
                np.zeros((100, 100)),
                [min_obs_plane_wavelength_in_nm, max_obs_plane_wavelength_in_nm],
                self.smbh_mass_exp,
                self.redshift_source,
                100,
                inclination_angle,
                OmM=self.OmM,
                H0=self.H0,
            )

        flux_map, _ = project_blr_to_source_plane(
            self.density_grid,
            self.z_velocity_grid,
            self.r_velocity_grid,
            inclination_angle,
            self.smbh_mass_exp,
            velocity_range=velocity_range,
            weighting_grid=self.emission_efficiency_array,
            radial_resolution=self.radial_step,
            vertical_resolution=self.height_step,
        )

        flux_map *= self.line_strength

        max_radius = (
            self.max_height * np.tan(inclination_angle * np.pi / 180) + self.max_radius
        )

        flux_projection = FluxProjection(
            flux_map,
            [min_obs_plane_wavelength_in_nm, max_obs_plane_wavelength_in_nm],
            self.smbh_mass_exp,
            self.redshift_source,
            max_radius,
            inclination_angle,
            OmM=self.OmM,
            H0=self.H0,
        )

        return flux_projection

    def calculate_blr_scattering_transfer_function(self, inclination_angle):
        """Calculate the transfer function of the BLR under the assumption of basic
        scattering by the particles. Essentially a density-weighted scattering with no
        wavelength dependence.

        :param inclination_angle: inclination of the source with respect to the observer
            in degrees
        :return: list representing the BLR's transfer function under the assumption that
            all particles can scatter photons of any wavelength with time units R_g / c
        """

        return calculate_blr_transfer_function(
            self.density_grid,
            self.z_velocity_grid,
            self.r_velocity_grid,
            inclination_angle,
            self.smbh_mass_exp,
            velocity_range=[-1, 1],
            weighting_grid=self.density_grid,
            radial_resolution=self.radial_step,
            vertical_resolution=self.height_step,
        )

    def calculate_blr_emission_line_transfer_function(
        self,
        inclination_angle,
        velocity_range=None,
        observed_wavelength_range_in_nm=None,
        speclite_filter=None,
        emission_efficiency_array=None,
    ):
        """Calculate the transfer function of the BLR under the assumption that the
        emission is related to the density * emission_efficiency_array. Unlike
        calculate_blr_scattering_transfer_function, this has wavelength dependence.
        Assumes the major time delay is related to the light travel time and the
        relaxation time of an excited particle is negligable.

        :param inclination_angle: angle of orientation of the source with respect to the
            observer in degrees.
        :param velocity_range: a list representing the min and max velocities to choose.
            Cannot be used with observed_wavelength_range_in_nm or speclite_filter. Note
            that positive velocity is toward the observer.
        :param observed_wavelength_range_in_nm: a list representing the min and max
            wavelengths in nanometers which are observed. Cannot be used with
            velocity:range or speclite_filter.
        :param speclite_filter: Speclite filter object or string representing a loadable
            Speclite filter. Cannot be used with velocity_range or
            observed_wavelength_range_in_nm.
        :param emission_efficiency_array: a 2d grid of weighting factors in the R-Z
            plane. Primarily used to selectively weight regions in the local optimally
            emitting cloud model.
        :return: list representing the BLR's transfer function under only a subset of
            the BLR particles can scatter photons in time units of R_g / c
        """

        if velocity_range is not None and observed_wavelength_range_in_nm is not None:
            print("Please only provide the velocities or wavelengths. Not both!")
            return False
        if (
            velocity_range is None
            and observed_wavelength_range_in_nm is None
            and speclite_filter is None
        ):
            print("Please provide the velocities or wavelengths.")
            return False
        if speclite_filter is not None:
            observed_wavelength_range_in_nm = (
                convert_speclite_filter_to_wavelength_range(
                    speclite_filter,
                )
            )
        if observed_wavelength_range_in_nm is not None:
            velocity_range = determine_emission_line_velocities(
                self.rest_frame_wavelength_in_nm,
                np.min(observed_wavelength_range_in_nm),
                np.max(observed_wavelength_range_in_nm),
                self.redshift_source,
            )

        expected_broadening = self.estimate_doppler_broadening(inclination_angle)
        obs_plane_wavelength_in_nm = self.rest_frame_wavelength_in_nm * (
            1 + self.redshift_source
        )
        min_obs_plane_wavelength_in_nm = (
            obs_plane_wavelength_in_nm
            * ((1 - np.max(velocity_range)) / (1 + np.max(velocity_range))) ** 0.5
        )
        max_obs_plane_wavelength_in_nm = (
            obs_plane_wavelength_in_nm
            * ((1 - np.min(velocity_range)) / (1 + np.min(velocity_range))) ** 0.5
        )

        min_expected_wavelength_in_nm = (
            obs_plane_wavelength_in_nm
            * ((1 - np.max(expected_broadening)) / (1 + np.max(expected_broadening)))
            ** 0.5
        )
        max_expected_wavelength_in_nm = (
            obs_plane_wavelength_in_nm
            * ((1 - np.min(expected_broadening)) / (1 + np.min(expected_broadening)))
            ** 0.5
        )

        if min_expected_wavelength_in_nm > max_obs_plane_wavelength_in_nm:
            return 0, np.zeros(3)
        if max_expected_wavelength_in_nm < min_obs_plane_wavelength_in_nm:
            return 0, np.zeros(3)

        self.set_emission_efficiency_array(emission_efficiency_array)

        emission_line_tf = calculate_blr_transfer_function(
            self.density_grid,
            self.z_velocity_grid,
            self.r_velocity_grid,
            inclination_angle,
            self.smbh_mass_exp,
            velocity_range=velocity_range,
            weighting_grid=self.emission_efficiency_array,
            radial_resolution=self.radial_step,
            vertical_resolution=self.height_step,
        )

        if self.current_calculated_inclination is not inclination_angle:
            self.current_calculated_inclination = inclination_angle
            self.emission_efficiency_array = emission_efficiency_array
            total_emission = self.project_blr_total_intensity(
                self.current_calculated_inclination,
                emission_efficiency_array=self.emission_efficiency_array,
            )
            self.current_total_emission = total_emission.total_flux

        current_emission = self.project_blr_intensity_over_velocity_range(
            inclination_angle,
            velocity_range=velocity_range,
            emission_efficiency_array=self.emission_efficiency_array,
        )

        tf_weighting_factor = current_emission.total_flux / self.current_total_emission
        tf_weighting_factor *= self.line_strength

        return tf_weighting_factor, emission_line_tf

    def estimate_doppler_broadening(self, inclination_angle):
        """Estimate the maximum and minimum velocities of the BLR using the maximum
        values stored in velocity arrays. This is a helper method which determines when
        it's worth integrating through the BLR and when the integration will return 0.

        :param inclination_angle: inclination of the source w.r.t the observer in
            degrees.
        :return: array containing the estimated maximum receeding and approaching
            velocities in units v / c
        """

        assert inclination_angle < 90
        assert inclination_angle >= 0

        inclination_angle *= np.pi / 180
        max_range_v_r = np.max(
            [abs(np.max(self.r_velocity_grid)), abs(np.min(self.r_velocity_grid))]
        )
        min_range_v_z = np.min(self.z_velocity_grid)
        max_range_v_z = np.max(self.z_velocity_grid)
        densities_as_function_of_radius = np.sum(self.density_grid, axis=1)
        min_radius_argument = np.argmax(
            np.isfinite(1 / densities_as_function_of_radius)
        )
        min_radius_in_rg = self.radii_values[min_radius_argument]

        max_phi_velocity = calculate_keplerian_velocity(
            min_radius_in_rg * self.rg, 10**self.smbh_mass_exp
        )

        max_receeding_velocity = np.cos(inclination_angle) * min_range_v_z - np.sin(
            inclination_angle
        ) * np.sqrt(max_range_v_r**2 + max_phi_velocity**2)
        max_approaching_velocity = np.cos(inclination_angle) * max_range_v_z + np.sin(
            inclination_angle
        ) * np.sqrt(max_range_v_r**2 + max_phi_velocity**2)

        return np.asarray([max_receeding_velocity, max_approaching_velocity])

    def update_line_strength(self, line_strength):
        """Update the emission line strength stored in the BLR

        :param line_strength: updated strength of the emisison line w.r.t. the
            continuum emission
        :return: True if successful
        """
        prev_line_strength = self.line_strength
        assert isinstance(line_strength, (int, float))
        self.line_strength = line_strength
        if self.current_total_emission is not None and prev_line_strength != 0:
            self.current_total_emission *= self.line_strength / prev_line_strength
        return True

    def get_density_axis(self):
        """Get a meshgrid representation of the R-Z coordinates to get an
        array to apply local optimally emitting cloud region weighting. The spherical
        distance may be computed as:

        r_spherical = np.sqrt(R**2 * Z**2)

        :return: R and Z coordinates in a numpy meshgrid
        """

        R, Z = np.meshgrid(self.radii_values, self.height_values, indexing="ij")
        return R, Z

    def set_emission_efficiency_array(self, emission_efficiency_array=None):
        """Set the weighting of each (R, Z) coordinate. This is used to compute the
        emission and response of the BLR for spatially distinct regions.

        :param emission_efficiency_array: Array of int/float values representing
            how efficient the BLR responds at coordinate (R, Z)
        :return: True if successful
        """
        R, Z = self.get_density_axis()
        if emission_efficiency_array is None:
            self.emission_efficiency_array = np.ones(np.shape(R))
        else:
            assert np.shape(emission_efficiency_array) == np.shape(R)
            self.emission_efficiency_array = emission_efficiency_array
        return True
